"""
Authentication-related structures.

Kopf handles some rudimentary authentication directly, and exposes the ways
to implement custom authentication methods (via ``on.login`` handlers).

For that, a minimally sufficient data structure is introduced --- both
to bring all the credentials together in a structured and type-annotated way,
and to receive them from the operators' login-handlers with custom auth methods.

The "rudimentary" is defined as the information passed to the HTTP protocol
and TCP/SSL connection only, i.e. everything usable in a generic HTTP client,
and nothing more than that:

* TCP server host & port.
* SSL verification/ignorance flag.
* SSL certificate authority.
* SSL client certificate and its private key.
* HTTP ``Authorization: Basic username:password``.
* HTTP ``Authorization: Bearer token`` (or other schemes: Bearer, Digest, etc).
* URL's default namespace for the cases when this is implied.

.. seealso::
    :func:`authentication` and :mod:`piggybacking`.
"""
import asyncio
import base64
import collections
import contextlib
import dataclasses
import datetime
import inspect
import os
import random
import ssl
import tempfile
from collections.abc import AsyncIterable, AsyncIterator, Callable
from typing import NewType, TypeVar, cast

import aiohttp


class LoginError(Exception):
    """ Raised when the operator cannot login to the API. """


class AccessError(Exception):
    """ Raised when the operator cannot access the cluster API. """


# Naming: "ConnectionInfo" is the best name, but it is already a part of the public interface.
# Anything "Cluster…" is too narrow. `Credentials` are more suited for the current `ConnectionInfo`.
# "KubeContext" mimics the kubeconfig's terminology & content, so seemingly fits the best.
# The class is anyway hidden from users, so we can use any name. Class inheritance is not supported.
@dataclasses.dataclass(frozen=True, kw_only=True)
class KubeContext:
    """
    A base connection info with no credentials (added in descendant classes).
    """
    server: str  # e.g. "https://localhost:443"
    priority: int = 0
    default_namespace: str | None = None  # used for cluster objects' k8s-events.
    expiration: datetime.datetime | None = None  # TZ-aware or TZ-naive (implies UTC)


@dataclasses.dataclass(frozen=True, kw_only=True)
class ConnectionInfo(KubeContext):
    """
    A single endpoint with specific credentials and connection flags to use.
    """
    server: str  # e.g. "https://localhost:443"
    ca_path: str | bytes | os.PathLike[str] | os.PathLike[bytes] | None = None
    ca_data: str | bytes | None = None
    insecure: bool | None = None
    username: str | None = None
    password: str | None = None
    scheme: str | None = None  # RFC-7235/5.1: e.g. Bearer, Basic, Digest, etc.
    token: str | None = None
    certificate_path: str | bytes | os.PathLike[str] | os.PathLike[bytes] | None = None
    certificate_data: str | bytes | None = None
    private_key_path: str | bytes | os.PathLike[str] | os.PathLike[bytes] | None = None
    private_key_data: str | bytes | None = None
    default_namespace: str | None = None  # used for cluster objects' k8s-events.
    proxy_url: str | None = None
    trust_env: bool = False
    priority: int = 0
    expiration: datetime.datetime | None = None  # TZ-aware or TZ-naive (implies UTC)

    def __post_init__(self) -> None:
        if self.ca_path and self.ca_data:
            raise ValueError("Both CA path & data are set. Need only one.")
        if self.certificate_path and self.certificate_data:
            raise ValueError("Both certificate path & data are set. Need only one.")
        if self.private_key_path and self.private_key_data:
            raise ValueError("Both private key path & data are set. Need only one.")

    def as_aiohttp_basic_auth(self) -> aiohttp.BasicAuth | None:
        """Make a basic auth for username/password, or ``None`` if absent."""
        if self.username and self.password:
            return aiohttp.BasicAuth(self.username, self.password)
        else:
            return None

    def as_http_headers(self) -> dict[str, str]:
        """
        Make a dict with the ``Authorization`` header set to scheme+token,
        or an empty dict if there are no tokens or schemes.
        """
        if self.scheme and self.token:
            return {'Authorization': f'{self.scheme} {self.token}'}
        elif self.scheme:
            return {'Authorization': f'{self.scheme}'}
        elif self.token:
            return {'Authorization': f'Bearer {self.token}'}
        else:
            return {}

    def as_ssl_context(self) -> ssl.SSLContext:
        """
        Make an SSL context with CA and client cert using Python's :mod:`ssl`.

        .. warning::
            It will store the :attr:`kopf.ConnectionInfo.certificate_data` and
            :attr:`kopf.ConnectionInfo.private_key_data` into temporary files
            for a brief moment of time until the SSL context is constructed,
            since Python's :mod:`ssl` cannot load them from memory.
        """
        # Some SSL data are not accepted directly, so we have to use temp files.
        # Do not even create temporary files if there is no need. It can be a readonly filesystem.
        with contextlib.ExitStack() as stack:

            cert_path: str | bytes | os.PathLike[str] | os.PathLike[bytes] | None
            if self.certificate_path:
                cert_path = self.certificate_path
            elif self.certificate_data:
                cert_file = stack.enter_context(tempfile.NamedTemporaryFile(buffering=0))
                cert_file.write(self.__decode_to_pem(self.certificate_data).encode('ascii'))
                cert_path = cert_file.name
            else:
                cert_path = None

            pkey_path: str | bytes | os.PathLike[str] | os.PathLike[bytes] | None
            if self.private_key_path:
                pkey_path = self.private_key_path
            elif self.private_key_data:
                pkey_file = stack.enter_context(tempfile.NamedTemporaryFile(buffering=0))
                pkey_file.write(self.__decode_to_pem(self.private_key_data).encode('ascii'))
                pkey_path = pkey_file.name
            else:
                pkey_path = None

            # The SSL part (both client certificate auth and CA verification).
            context: ssl.SSLContext
            if cert_path and pkey_path:
                context = ssl.create_default_context(
                    purpose=ssl.Purpose.SERVER_AUTH,
                    cafile=self.ca_path,
                    cadata=self.__decode_to_pem(self.ca_data) if self.ca_data is not None else None,
                )
                context.load_cert_chain(certfile=cert_path, keyfile=pkey_path)
            else:
                context = ssl.create_default_context(
                    cafile=self.ca_path,
                    cadata=self.__decode_to_pem(self.ca_data) if self.ca_data is not None else None,
                )

        if self.insecure:
            context.check_hostname = False
            context.verify_mode = ssl.CERT_NONE

        return context

    @staticmethod
    def __decode_to_pem(data: str | bytes) -> str:
        match data:
            case str() if data.startswith('-----BEGIN '):
                return data
            case bytes() if data.startswith(b'-----BEGIN '):
                return data.decode('ascii')
            case _:
                return base64.b64decode(data).decode('ascii')


@dataclasses.dataclass(frozen=True, kw_only=True)
class AiohttpSession(KubeContext):
    """
    A custom ``aiohttp`` session to use instead of the built-in one.

    See: :ref:`custom-http-sessions` for details.
    """
    aiohttp_session: aiohttp.ClientSession


_T = TypeVar('_T', bound=object)

# Usually taken from the HandlerId (also a string), but semantically it is on its own.
VaultKey = NewType('VaultKey', str)


@dataclasses.dataclass
class VaultItem:
    """
    The actual item stored in the vault. It is never exposed externally.

    Used for proper garbage collection when the key is removed from the vault
    (to avoid orchestrating extra cache structures and keeping them in sync).

    The caches are populated by :meth:`Vault.extended` on-demand.
    """
    info: KubeContext
    caches: dict[str, object] | None = None


class Vault(AsyncIterable[tuple[VaultKey, KubeContext]]):
    """
    A store for currently valid authentication methods.

    *Though we call it a vault to add a sense of security.*

    Normally, only one authentication method is used at a time in multiple
    methods and tasks (e.g. resource watching/patching, peering, etc.).

    Multiple methods to represent the same principal is an unusual case,
    but it is also possible as a side effect. Same for multiple distinct
    identities of a single operator.

    The credentials store is created once for an operator (a task),
    and is then used by multiple tasks running in parallel:

    * Consumed by the API client wrappers to authenticate in the API.
    * Reported by the API client wrappers if some of the credentials fail.
    * Populated by the authenticator background task when and if needed.

    .. seealso::
        :func:`auth.authenticated` and :func:`authentication`.
    """
    _current: dict[VaultKey, VaultItem]
    _invalid: dict[VaultKey, list[VaultItem]]

    def __init__(
            self,
            __src: dict[str, object] | None = None,
    ) -> None:
        super().__init__()
        self._current = {}
        self._invalid = collections.defaultdict(list)
        self._next_expiration: datetime.datetime | None = None

        if __src is not None:
            self._update_converted(__src)

        # Mark a pre-populated vault to be usable instantly,
        # or trigger the initial authentication for an empty vault.
        self._guard = asyncio.Condition()
        self._ready: bool = not self.is_empty()

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}: {self._current!r}>'

    def __bool__(self) -> bool:
        raise NotImplementedError("The vault should not be evaluated as bool.")

    async def __aiter__(
            self,
    ) -> AsyncIterator[tuple[VaultKey, KubeContext]]:
        async for key, item in self._items():
            yield key, item.info

    async def extended(
            self,
            factory: Callable[[KubeContext], _T],
            purpose: str | None = None,
    ) -> AsyncIterator[tuple[VaultKey, KubeContext, _T]]:
        """
        Iterate the connection info items with their cached object.

        The cached objects are identified by the purpose (an arbitrary string).
        Multiple types of objects can be cached under different names.

        The factory is a one-argument function of :class:`.KubeContext`,
        that returns the object to be cached for this connection info.
        It is called only once per item and purpose.
        """
        purpose = purpose if purpose is not None else repr(factory)
        async for key, item in self._items():
            if item.caches is None:  # quick-check with no locking overhead.
                async with self._guard:
                    if item.caches is None:  # securely synchronised check.
                        item.caches = {}
            if purpose not in item.caches:  # quick-check with no locking overhead.
                async with self._guard:
                    if purpose not in item.caches:  # securely synchronised check.
                        item.caches[purpose] = factory(item.info)
            yield key, item.info, cast(_T, item.caches[purpose])

    async def _items(
            self,
    ) -> AsyncIterator[tuple[VaultKey, VaultItem]]:
        """
        Yield the raw items as stored in the vault in random order.

        The items are yielded until either all of them are depleted,
        or until the yielded one does not fail (no ``.invalidate`` call made).
        Restart on every re-authentication (if new items are added).
        """

        # Yield the connection infos until either all of them are depleted,
        # or until the yielded one does not fail (no `.invalidate` call made).
        # Restart on every re-authentication (if new items are added).
        while True:

            async with self._guard:

                # Whether on the 1st run, or during the active re-authentication,
                # ensure that the items are ready before yielding them.
                await self._guard.wait_for(lambda: self._ready)

                # Check for expiration strictly after a possible re-authentication.
                # This might cause another re-authentication if the credentials are pre-expired.
                await self._expire()

                # Select the items to yield and let it (i.e. a consumer task) work.
                yielded_key, yielded_item = self.select()

            # Yield strictly outside of locks/conditions. The vault must be free for invalidations.
            yield yielded_key, yielded_item

            # If the yielded item has been invalidated, assume that this item has failed.
            # Otherwise (the item is in the list), it has succeeded -- we are done iterating.
            # Note: checked by identity, in case a similar item is re-added as a different object.
            async with self._guard:
                if yielded_key in self._current and self._current[yielded_key] is yielded_item:
                    break

    def select(self) -> tuple[VaultKey, VaultItem]:
        """
        Select the next item (not the info!) to try (and do so infinitely).

        .. warning::
            This method is not async/await-safe: if the data change on the go,
            it can lead to improper items returned.
        """
        if not self._current:
            raise LoginError("Ran out of valid credentials. Consider installing "
                             "an API client library or adding a login handler. See more: "
                             "https://docs.kopf.dev/en/stable/authentication/")
        prioritised: dict[int, list[tuple[VaultKey, VaultItem]]]
        prioritised = collections.defaultdict(list)
        for key, item in self._current.items():
            prioritised[item.info.priority].append((key, item))
        top_priority = max(list(prioritised.keys()))
        key, item = random.choice(prioritised[top_priority])
        return key, item

    async def expire(self) -> None:  # unused, but declared public (deprecate)
        """
        Discard the expired credentials, and re-authenticate as needed.

        Unlike invalidation, the expired credentials are not remembered
        and not blocked from reappearing.
        """
        # Quick & lockless for speed: it is done on every API call, we have no time for locks.
        now = datetime.datetime.now(datetime.timezone.utc)
        if self._next_expiration is not None and now >= self._next_expiration:
            async with self._guard:
                await self._expire()

    async def _expire(self) -> None:
        """
        Discard the expired credentials, and re-authenticate as needed.

        Unlike invalidation, the expired credentials are not remembered
        and not blocked from reappearing.
        """
        now = datetime.datetime.now(datetime.timezone.utc)

        # Avoid waiting for re-auth afterwards if there is nothing to expire or change.
        expired = False
        if self._next_expiration is not None and now >= self._next_expiration:
            for key, item in list(self._current.items()):
                expiration = item.info.expiration
                if expiration is not None:
                    if expiration.tzinfo is None:
                        expiration = expiration.replace(tzinfo=datetime.timezone.utc)
                    if now >= expiration:
                        await self._flush_caches(item)
                        del self._current[key]
                        expired = True
            self._update_expiration()

        # Initiate a re-authentication activity, and block until it is finished.
        if expired and not self._current:  # i.e. nothing is left at all
            self._ready = False
            self._guard.notify_all()
            await self._guard.wait_for(lambda: self._ready)

    async def invalidate(
            self,
            key: VaultKey,
            info: KubeContext,
            *,
            exc: Exception | None = None,
    ) -> None:
        """
        Discard the specified credentials, and re-authenticate as needed.

        Multiple calls can be made for a single authenticator and credentials,
        if used for multiple requests at the same time (a common case).
        All of them will be blocked the same way, until one and only one
        re-authentication happens in a background task. They will be
        unblocked at the same instant once the new credentials are ready.

        If the re-authentication fails in the background task, this method
        re-raises the original exception (most likely a HTTP 401 error),
        and lets the client tasks fail in their own stack.
        The background task continues to run and tries to re-authenticate
        on the next API calls until cancelled due to the operator exit.
        """
        # Exclude the failed connection items from the list of available ones.
        # But keep a short history of invalid items, so that they are not re-added.
        # The history size is estimated by the number of parallel streams trying to re-auth at once.
        async with self._guard:
            # Note: not "==", but "is". If not the same, then it was invalidated by other consumers,
            # the new current credentials is something new to use (maybe equal to the old one).
            if key in self._current and self._current[key].info is info:
                await self._flush_caches(self._current[key])
                self._invalid[key] = self._invalid[key][-2:] + [self._current[key]]
                del self._current[key]
                self._update_expiration()

            # Initiate a re-authentication activity, and block until it is finished.
            if not self._current:  # i.e. nothing is left at all
                self._ready = False
                self._guard.notify_all()
                await self._guard.wait_for(lambda: self._ready)

            # If the re-auth has failed, re-raise the original exception in the current stack.
            # If the original exception is unknown, raise normally on the next iteration's yield.
            # The error here is optional -- for better stack traces of the original exception `exc`.
            # Keep in mind, this routine is called in parallel from many tasks for the same keys.
            if not self._current:
                if exc is not None:
                    raise LoginError("Ran out of valid credentials. Consider installing "
                                     "an API client library or adding a login handler. See more: "
                                     "https://docs.kopf.dev/en/stable/authentication/") from exc

    async def populate(
            self,
            __src: dict[str, object],
    ) -> None:
        """
        Add newly retrieved credentials.

        Used by :func:`authentication` to add newly retrieved credentials
        from the authentication activity handlers. Some of the credentials
        can be duplicates of the existing ones --- only one of them is used then.
        """
        async with self._guard:

            # Remember the new credentials or replace the old ones. If we already see that the item
            # is invalid (as seen in our short per-key history), we keep it as such -- this prevents
            # repeatedly invalid credentials from causing infinite re-authentication again & again.
            self._update_converted(__src)

            # Notify the consuming tasks (API clients) that new credentials are ready to be used.
            # Those tasks can be blocked in `vault.invalidate()` if there are no credentials left.
            self._ready = True
            self._guard.notify_all()

    def is_empty(self) -> bool:
        now = datetime.datetime.now(datetime.timezone.utc)
        expirations = [
            dt if dt is None or dt.tzinfo is not None else dt.replace(tzinfo=datetime.timezone.utc)
            for dt in (item.info.expiration for item in self._current.values())
        ]
        return all(dt is not None and now >= dt for dt in expirations)  # i.e. expired

    async def wait_for_readiness(self) -> None:
        async with self._guard:
            await self._guard.wait_for(lambda: self._ready)

    async def wait_for_emptiness(self) -> None:
        async with self._guard:
            await self._guard.wait_for(lambda: not self._ready)

    async def close(self) -> None:
        """
        Finalize all the cached objects when the operator is ending.
        """
        async with self._guard:
            for key in self._current:
                await self._flush_caches(self._current[key])

    async def _flush_caches(
            self,
            item: VaultItem,
    ) -> None:
        """
        Call the finalizers and garbage-collect the cached objects.

        Mainly used to garbage-collect aiohttp sessions and its derivatives
        when the connection info items are removed from the vault --- so that
        the sessions/connectors would not complain that they were not closed.

        Built-in garbage-collection is not sufficient, as it is synchronous, and
        cannot call the async coroutines like ``aiohttp.ClientSession.close()``.

        .. note::
            Currently, we assume the ``close()`` method only (both sync/async).
            There is no need to generalise to customizable finalizer callbacks.
            This can change in the future.
        """

        # Close the closable objects.
        if item.caches:
            for obj in item.caches.values():
                if hasattr(obj, 'close'):
                    if inspect.iscoroutinefunction(getattr(obj, 'close')):
                        await getattr(obj, 'close')()
                    else:
                        getattr(obj, 'close')()

        # Garbage-collect other resources (e.g. files, memory, etc).
        item.caches = None

    def _update_converted(
            self,
            __src: dict[str, object],
    ) -> None:
        for key, info in __src.items():
            key = VaultKey(str(key))
            if not isinstance(info, KubeContext):
                raise ValueError("Only ConnectionInfo/AiohttpSession instances are accepted.")
            if info not in [data.info for data in self._invalid[key]]:
                self._current[key] = VaultItem(info=info)
        self._update_expiration()

    def _update_expiration(self) -> None:
        expirations = [
            dt if dt.tzinfo is not None else dt.replace(tzinfo=datetime.timezone.utc)
            for dt in (item.info.expiration for item in self._current.values())
            if dt is not None
        ]
        self._next_expiration = min(expirations) if expirations else None



import kopf
from kopf._core.actions.execution import handler_var
from kopf._core.actions.invocation import context
from kopf._core.reactor.subhandling import subregistry_var


# Used in the tests. Must be global-scoped, or its qualname will be affected.
def child_fn(**_):
    pass


def test_with_parent(
        parent_handler, resource_registry_cls, cause_factory):

    cause = cause_factory(resource_registry_cls)
    registry = resource_registry_cls()

    with context([(handler_var, parent_handler), (subregistry_var, registry)]):
        kopf.subhandler()(child_fn)

    handlers = registry.get_handlers(cause)
    assert len(handlers) == 1
    assert handlers[0].fn is child_fn
    assert handlers[0].id == 'parent_fn/child_fn'



import pytest


@pytest.fixture(autouse=True)
def _enforced_api_server(fake_vault, enforced_session, resource):
    pass


@pytest.fixture(autouse=True)
def _prevent_retries_in_api_tests(settings):
    settings.networking.error_backoffs = []



"""
The test design notes:

* The field "abc.def.hij" is existent.
* The field "rst.uvw.xyz" is inexistent.
* Its mixtures ("a.b.z", "a.y.z") are used to simulate partial inexistence.

* For the existent keys, kwargs should not matter.
* For the non-existent keys, the default is returned,
  or a ``KeyError`` raised --- as with regular mappings.

* For special cases with "wrong" values (``"value"["z"]``, ``None["z"]``, etc),
  either a ``TypeError`` should be raised normally. If "wrong" values are said
  to be ignored, then they are treated the same as inexistent values,
  and the default value is returned or a ``KeyError`` is raised.

"""
import types

import pytest

from kopf._cogs.structs.dicts import resolve, resolve_obj

default = object()


#
# Both resolve functions should behave exactly the same for dicts.
#
@pytest.mark.parametrize('resolve', [resolve, resolve_obj])
def test_dict_with_existent_key_with_no_default(resolve):
    d = {'abc': {'def': {'hij': 'val'}}}
    r = resolve(d, ['abc', 'def', 'hij'])
    assert r == 'val'


@pytest.mark.parametrize('resolve', [resolve, resolve_obj])
def test_dict_with_existent_key_with_default(resolve):
    d = {'abc': {'def': {'hij': 'val'}}}
    r = resolve(d, ['abc', 'def', 'hij'], default)
    assert r == 'val'


@pytest.mark.parametrize('key', [
    pytest.param(['rst', 'uvw', 'xyz'], id='1stlvl'),
    pytest.param(['abc', 'uvw', 'xyz'], id='2ndlvl'),
    pytest.param(['abc', 'def', 'xyz'], id='3rdlvl'),
])
@pytest.mark.parametrize('resolve', [resolve, resolve_obj])
def test_dict_with_inexistent_key_with_no_default(resolve, key):
    d = {'abc': {'def': {'hij': 'val'}}}
    with pytest.raises(KeyError):
        resolve(d, key)


@pytest.mark.parametrize('key', [
    pytest.param(['rst', 'uvw', 'xyz'], id='1stlvl'),
    pytest.param(['abc', 'uvw', 'xyz'], id='2ndlvl'),
    pytest.param(['abc', 'def', 'xyz'], id='3rdlvl'),
])
@pytest.mark.parametrize('resolve', [resolve, resolve_obj])
def test_dict_with_inexistent_key_with_default(resolve, key):
    d = {'abc': {'def': {'hij': 'val'}}}
    r = resolve(d, key, default)
    assert r is default


@pytest.mark.parametrize('resolve', [resolve, resolve_obj])
def test_dict_with_nonmapping_with_no_default(resolve):
    d = {'key': 'val'}
    with pytest.raises(TypeError):
        resolve(d, ['key', 'sub'])


@pytest.mark.parametrize('resolve', [resolve, resolve_obj])
def test_dict_with_nonmapping_with_default(resolve):
    d = {'key': 'val'}
    r = resolve(d, ['key', 'sub'], default)
    assert r is default


@pytest.mark.parametrize('resolve', [resolve, resolve_obj])
def test_dict_with_none_is_treated_as_a_regular_default_value(resolve):
    d = {'abc': {'def': {'hij': 'val'}}}
    r = resolve(d, ['abc', 'def', 'xyz'], None)
    assert r is None


@pytest.mark.parametrize('resolve', [resolve, resolve_obj])
def test_dict_with_empty_path(resolve):
    d = {'key': 'val'}
    r = resolve(d, [])
    assert r == d
    assert r is d


#
# Specialised drill-down for objects.
#
class FakeKubernetesModel:  # no bases!
    __module__ = 'kubernetes.client.models.fake-for-tests'

    @property
    def metadata(self):
        return None

    attribute_map = {
        'AbC': 'abc',
        'zzz': 'dez',
    }


@pytest.fixture(params=[FakeKubernetesModel, types.SimpleNamespace])
def obj(request):
    cls = request.param
    obj = cls()
    if cls is FakeKubernetesModel:
        # With attribute mapping in mind.
        obj.key = 'val'
        obj.AbC = cls()
        obj.AbC.zzz = cls()
        obj.AbC.zzz.hij = 'val'
    else:
        # Exactly as they will be requested.
        obj.key = 'val'
        obj.abc = cls()
        obj.abc.dez = cls()
        obj.abc.dez.hij = 'val'
    return obj


def test_object_with_existent_key_with_no_default(obj):
    r = resolve_obj(obj, ['abc', 'dez', 'hij'])
    assert r == 'val'


def test_object_with_existent_key_with_default(obj):
    r = resolve_obj(obj, ['abc', 'dez', 'hij'], default)
    assert r == 'val'


@pytest.mark.parametrize('key', [
    pytest.param(['rst', 'uvw', 'xyz'], id='1stlvl'),
    pytest.param(['abc', 'uvw', 'xyz'], id='2ndlvl'),
    pytest.param(['abc', 'dez', 'xyz'], id='3rdlvl'),
])
def test_object_with_inexistent_key_with_no_default(obj, key):
    with pytest.raises(AttributeError):
        resolve_obj(obj, key)


@pytest.mark.parametrize('key', [
    pytest.param(['rst', 'uvw', 'xyz'], id='1stlvl'),
    pytest.param(['abc', 'uvw', 'xyz'], id='2ndlvl'),
    pytest.param(['abc', 'dez', 'xyz'], id='3rdlvl'),
])
def test_object_with_inexistent_key_with_default(obj, key):
    r = resolve_obj(obj, key, default)
    assert r is default


def test_object_with_nonmapping_with_no_default(obj):
    with pytest.raises(TypeError):
        resolve_obj(obj, ['key', 'sub'])


def test_object_with_nonmapping_with_default(obj):
    r = resolve_obj(obj, ['key', 'sub'], default)
    assert r is default


def test_object_with_none_is_treated_as_a_regular_default_value(obj):
    r = resolve_obj(obj, ['abc', 'dez', 'xyz'], None)
    assert r is None


def test_object_with_empty_path(obj):
    r = resolve_obj(obj, [])
    assert r == obj
    assert r is obj


#
# Some special cases.
#
@pytest.mark.parametrize('cls', (tuple, list, set, frozenset, str, bytes))
def test_raises_for_builtins(cls):
    obj = cls()
    with pytest.raises(TypeError):
        resolve_obj(obj, ['__class__'])



import asyncio

from kopf._cogs.clients.watching import Bookmark, continuous_watch


async def test_listed_is_inbetween(settings, resource, namespace, kmock):

    # Resource version is used as a continuation for the watch-queries.
    kmock['list', resource, kmock.namespace(namespace)] << {
        'metadata': {'resourceVersion': '123'},
        'items': [{'spec': 'a'}, {'spec': 'b'}],
    }

    # The same as in the `stream` fixture. But here, we also mock lists.
    kmock['watch', resource, kmock.namespace(namespace), kmock.params(resourceVersion='123')] << (
        {'type': 'ADDED', 'object': {'spec': 'c'}},
        {'type': 'ADDED', 'object': {'spec': 'd'}},
        {'type': 'ERROR', 'object': {'code': 410}},
    )

    events = []
    async for event in continuous_watch(settings=settings,
                                        resource=resource,
                                        namespace=namespace,
                                        operator_pause_waiter=asyncio.Future()):
        events.append(event)

    assert len(events) == 5
    assert events[0]['object']['spec'] == 'a'
    assert events[1]['object']['spec'] == 'b'
    assert events[2] == Bookmark.LISTED
    assert events[3]['object']['spec'] == 'c'
    assert events[4]['object']['spec'] == 'd'


async def test_bookmark_event_is_yielded(settings, resource, namespace, kmock):

    kmock['list', resource, kmock.namespace(namespace)] << {
        'metadata': {'resourceVersion': '100'},
        'items': [],
    }
    kmock['watch', resource, kmock.namespace(namespace), kmock.params(resourceVersion='100')] << (
        {'type': 'ADDED', 'object': {'metadata': {'resourceVersion': '101'}}},
        {'type': 'BOOKMARK', 'object': {'metadata': {'resourceVersion': '102'}}},
        {'type': 'ADDED', 'object': {'metadata': {'resourceVersion': '103'}}},
        {'type': 'ERROR', 'object': {'code': 410}},
    )

    events = []
    async for event in continuous_watch(settings=settings,
                                        resource=resource,
                                        namespace=namespace,
                                        operator_pause_waiter=asyncio.Future()):
        events.append(event)

    assert len(events) == 4
    assert events[0] == Bookmark.LISTED
    assert events[1]['type'] == 'ADDED'
    assert events[2]['type'] == 'BOOKMARK'
    assert events[3]['type'] == 'ADDED'


async def test_bookmark_updates_resource_version(settings, resource, namespace, kmock):

    kmock['list', resource, kmock.namespace(namespace)] << {
        'metadata': {'resourceVersion': '100'},
        'items': [],
    }

    # First watch: a bookmark advances the resource version, then the stream ends naturally.
    kmock['watch', resource, kmock.namespace(namespace), kmock.params(resourceVersion='100')] << (
        {'type': 'BOOKMARK', 'object': {'metadata': {'resourceVersion': '200'}}},
    )

    # Second watch: must resume from the bookmark's resource version, not the list's.
    kmock['watch', resource, kmock.namespace(namespace), kmock.params(resourceVersion='200')] << (
        {'type': 'ADDED', 'object': {'metadata': {'resourceVersion': '201'}}},
        {'type': 'ERROR', 'object': {'code': 410}},
    )

    events = []
    async for event in continuous_watch(settings=settings,
                                        resource=resource,
                                        namespace=namespace,
                                        operator_pause_waiter=asyncio.Future()):
        events.append(event)

    assert len(events) == 3
    assert events[0] == Bookmark.LISTED
    assert events[1]['type'] == 'BOOKMARK'
    assert events[2]['type'] == 'ADDED'



from collections.abc import Collection

from kopf._cogs.clients import api
from kopf._cogs.configs import configuration
from kopf._cogs.helpers import typedefs
from kopf._cogs.structs import bodies, references


async def list_objs(
        *,
        settings: configuration.OperatorSettings,
        resource: references.Resource,
        namespace: references.Namespace,
        logger: typedefs.Logger,
) -> tuple[Collection[bodies.RawBody], str]:
    """
    List the objects of specific resource type.

    The cluster-scoped call is used in two cases:

    * The resource itself is cluster-scoped, and namespacing makes no sense.
    * The operator serves all namespaces for the namespaced custom resource.

    Otherwise, the namespace-scoped call is used:

    * The resource is namespace-scoped AND operator is namespaced-restricted.
    """
    rsp = await api.get(
        url=resource.get_url(namespace=namespace),
        logger=logger,
        settings=settings,
    )

    items: list[bodies.RawBody] = []
    resource_version = rsp.get('metadata', {}).get('resourceVersion', None)
    for item in rsp.get('items', []):
        if 'kind' in rsp:
            item.setdefault('kind', rsp['kind'].removesuffix('List'))
        if 'apiVersion' in rsp:
            item.setdefault('apiVersion', rsp['apiVersion'])
        items.append(item)

    return items, resource_version



import asyncio

import pytest

import kopf

# We assume that the handler filtering is tested in details elsewhere (for all handlers).
# Here, we only test if it is applied or not applied.


async def test_timer_filtration_satisfied(
        settings, resource, dummy, assert_logs, k8s_mocked, simulate_cycle):
    executed = asyncio.Event()

    @kopf.timer(*resource, id='fn',
                labels={'a': 'value', 'b': kopf.PRESENT, 'c': kopf.ABSENT},
                annotations={'x': 'value', 'y': kopf.PRESENT, 'z': kopf.ABSENT})
    async def fn(**kwargs):
        dummy.mock(**kwargs)
        executed.set()

    event_body = {'metadata': {'labels': {'a': 'value', 'b': '...'},
                               'annotations': {'x': 'value', 'y': '...'},
                               'finalizers': [settings.persistence.finalizer]}}
    await simulate_cycle(event_body)
    await executed.wait()

    assert dummy.mock.call_count == 1


@pytest.mark.parametrize('labels, annotations', [
    # Annotations mismatching (but labels are matching):
    ({'a': 'value', 'b': '...'}, {'x': 'mismatching-value', 'b': '...'}, ),  # x must be "value".
    ({'a': 'value', 'b': '...'}, {'x': 'value', 'y': '...', 'z': '...'}),  # z must be absent
    ({'a': 'value', 'b': '...'}, {'x': 'value'}),  # y must be present
    # labels mismatching (but annotations are matching):
    ({'a': 'mismatching-value', 'b': '...'}, {'x': 'value', 'y': '...'}),
    ({'a': 'value', 'b': '...', 'c': '...'}, {'x': 'value', 'y': '...'}),
    ({'a': 'value'}, {'x': 'value', 'y': '...'}),
])
async def test_timer_filtration_mismatched(
        settings, resource, mocker, labels, annotations,
        assert_logs, k8s_mocked, simulate_cycle):
    spawn_daemons = mocker.patch('kopf._core.engines.daemons.spawn_daemons')

    @kopf.timer(*resource, id='fn',
                labels={'a': 'value', 'b': kopf.PRESENT, 'c': kopf.ABSENT},
                annotations={'x': 'value', 'y': kopf.PRESENT, 'z': kopf.ABSENT})
    async def fn(**kwargs):
        pass

    event_body = {'metadata': {'labels': labels,
                               'annotations': annotations,
                               'finalizers': [settings.persistence.finalizer]}}
    await simulate_cycle(event_body)
    await asyncio.sleep(123)  # give it enough time to do something when nothing is expected

    assert spawn_daemons.called
    assert spawn_daemons.call_args_list[0].kwargs['handlers'] == []



"""
Rudimentary type [re-]definitions for cross-versioned Python & mypy.

The problem is that new mypy versions often bring type-sheds with StdLib types
defined as generics, while the old Python runtime (down to 3.9 & 3.10)
does not support the usual syntax.
Examples: asyncio.Task, asyncio.Future, logging.LoggerAdapter, and others.

This module defines them in a most suitable and reusable way. Plus it adds
some common plain type definitions used across the codebase (for convenience).
"""
import logging
from typing import TYPE_CHECKING, Any, TypeAlias

if TYPE_CHECKING:
    LoggerAdapter: TypeAlias = logging.LoggerAdapter[Any]
else:
    LoggerAdapter: TypeAlias = logging.LoggerAdapter

# As publicly exposed: we only promise that it is based on one of the built-in loggable classes.
# Mind that these classes have multi-versioned stubs, so we avoid redefining the protocol ourselves.
Logger: TypeAlias = logging.Logger | LoggerAdapter



from kopf._core.reactor.running import operator


async def test_spawn_tasks_result_passed_to_run_tasks(mocker):
    sentinel = object()
    spawn = mocker.patch('kopf._core.reactor.running.spawn_tasks', return_value=sentinel)
    run = mocker.patch('kopf._core.reactor.running.run_tasks')

    await operator(clusterwide=True)

    assert spawn.await_count == 1
    assert run.await_count == 1
    assert run.call_args.args[0] is sentinel


async def test_existing_tasks_passed_as_ignored(mocker):
    existing = frozenset()
    mocker.patch('kopf._cogs.aiokits.aiotasks.all_tasks', return_value=existing)
    mocker.patch('kopf._core.reactor.running.spawn_tasks', return_value=[])
    run = mocker.patch('kopf._core.reactor.running.run_tasks')

    await operator(clusterwide=True)

    assert run.call_args.kwargs['ignored'] is existing


async def test_kwargs_forwarded_to_spawn_tasks(mocker):
    spawn = mocker.patch('kopf._core.reactor.running.spawn_tasks', return_value=[])
    mocker.patch('kopf._core.reactor.running.run_tasks')

    await operator(
        clusterwide=True,
        priority=100,
        peering_name='my-peering',
        standalone=True,
        liveness_endpoint="/health",
    )

    kwargs = spawn.call_args.kwargs
    assert kwargs['clusterwide'] is True
    assert kwargs['priority'] == 100
    assert kwargs['peering_name'] == 'my-peering'
    assert kwargs['standalone'] is True
    assert kwargs['liveness_endpoint'] == "/health"



from typing import Any

import pytest

from kopf._cogs.clients.auth import APIContext, authenticated


def test_package_version():
    import kopf
    assert hasattr(kopf, '__version__')
    assert kopf.__version__  # not empty, not null


@pytest.mark.parametrize('version, useragent', [
    ('1.2.3', 'kopf/1.2.3'),
    ('1.2rc', 'kopf/1.2rc'),
    (None, 'kopf/unknown'),
])
async def test_http_user_agent_version(kmock, fake_vault, mocker, version, useragent):
    mocker.patch('kopf._cogs.helpers.versions.version', version)
    kmock['/'] << (lambda req: dict(req.headers))

    @authenticated
    async def get_it(url: str, *, context: APIContext) -> dict[str, Any]:
        response = await context.session.get(url)
        return await response.json()

    returned_headers = await get_it(str(kmock.url))
    assert returned_headers['User-Agent'] == useragent
    await fake_vault.close()  # to prevent ResourceWarnings for unclosed connectors



import enum

from kopf._cogs.aiokits import aioenums


class DaemonStoppingReason(enum.Flag):
    """
    A reason or combination of reasons for a daemon being terminated.

    Daemons are signalled to exit usually for two reasons: the operator itself
    is exiting or restarting, so all daemons of all resources must stop;
    or the individual resource was deleted, but the operator continues running.

    No matter the reason, the daemons must exit, so one and only one stop-flag
    is used. Some daemons can check the reason for exiting if it is important.

    There can be multiple reasons combined (in rare cases, all of them).
    """
    DONE = enum.auto()  # whatever the reason and the status, the asyncio task has exited.
    FILTERS_MISMATCH = enum.auto()  # the resource does not match the filters anymore.
    RESOURCE_DELETED = enum.auto()  # the resource was deleted, the asyncio task is still awaited.
    OPERATOR_PAUSING = enum.auto()  # the operator is pausing, the asyncio task is still awaited.
    OPERATOR_EXITING = enum.auto()  # the operator is exiting, the asyncio task is still awaited.
    DAEMON_SIGNALLED = enum.auto()  # the stopper flag was set, the asyncio task is still awaited.
    DAEMON_CANCELLED = enum.auto()  # the asyncio task was cancelled, the thread can be running.
    DAEMON_ABANDONED = enum.auto()  # we gave up on the asyncio task, the thread can be running.


DaemonStopper = aioenums.FlagSetter[DaemonStoppingReason]
DaemonStopped = aioenums.FlagWaiter[DaemonStoppingReason]
SyncDaemonStopperChecker = aioenums.SyncFlagWaiter[DaemonStoppingReason]  # deprecated
AsyncDaemonStopperChecker = aioenums.AsyncFlagWaiter[DaemonStoppingReason]  # deprecated



from kopf._cogs.structs.patches import Patch

# === Patch.__init__ ===


def test_patch_init_no_args():
    patch = Patch()
    assert dict(patch) == {}
    assert len(patch.fns) == 0


def test_patch_init_none_src():
    patch = Patch(None)
    assert dict(patch) == {}
    assert len(patch.fns) == 0


def test_patch_init_body_only():
    body = {'metadata': {'name': 'obj'}}
    patch = Patch(body=body)
    assert dict(patch) == {}
    assert len(patch.fns) == 0


def test_patch_init_body_stored_for_json_patch():
    body = {'abc': 456}
    patch = Patch(body=body)
    patch['abc'] = 789
    ops = patch.as_json_patch()
    assert ops == [{'op': 'replace', 'path': '/abc', 'value': 789}]


# === Patch.__bool__ ===


def test_patch_bool_empty_is_falsy():
    patch = Patch()
    assert not patch


def test_patch_bool_with_dict_is_truthy():
    patch = Patch({'a': 'b'})
    assert patch


def test_patch_bool_with_fns_only_is_truthy():
    patch = Patch(fns=[lambda body: None])
    assert patch


def test_patch_bool_with_dict_and_fns_is_truthy():
    patch = Patch({'a': 'b'}, fns=[lambda body: None])
    assert patch


# === Patch.__repr__ ===


def test_patch_repr_empty():
    patch = Patch()
    assert repr(patch) == "Patch()"


def test_patch_repr_dict_only():
    patch = Patch({'a': 'b'})
    assert repr(patch) == "Patch({'a': 'b'})"


def test_patch_repr_fns_only():
    patch = Patch(fns=[lambda body: None])
    r = repr(patch)
    assert r.startswith("Patch(fns=[")
    assert r.endswith("])")


def test_patch_repr_dict_and_fns():
    patch = Patch({'a': 'b'}, fns=[lambda body: None])
    r = repr(patch)
    assert r.startswith("Patch({'a': 'b'}, fns=[")
    assert r.endswith("])")


# === Patch.fns property and inheritance ===


def test_patch_fns_default_empty():
    patch = Patch()
    assert len(patch.fns) == 0


def test_patch_fns_from_constructor():
    fn1 = lambda body: None
    fn2 = lambda body: None
    patch = Patch(fns=[fn1, fn2])
    assert list(patch.fns) == [fn1, fn2]


def test_patch_inherits_fns_from_src_patch():
    fn1 = lambda body: None
    p1 = Patch(fns=[fn1])
    p2 = Patch(p1)
    assert list(p2.fns) == [fn1]


def test_patch_combines_inherited_and_own_fns():
    fn1 = lambda body: None
    fn2 = lambda body: None
    p1 = Patch(fns=[fn1])
    p2 = Patch(p1, fns=[fn2])
    assert list(p2.fns) == [fn1, fn2]


def test_patch_no_fns_from_plain_dict():
    p = Patch({'a': 'b'})
    assert len(p.fns) == 0


def test_patch_inherits_dict_and_fns():
    fn1 = lambda body: None
    p1 = Patch({'x': 'y'}, fns=[fn1])
    p2 = Patch(p1)
    assert p2['x'] == 'y'
    assert list(p2.fns) == [fn1]



#!/usr/bin/env python3
"""
Rudimentary HTTP CONNECT proxy server for tunneling HTTPS connections.
Preserves client certificates by creating a transparent TCP tunnel.

This is a testing tool itself, not a part of the package, so it is not tested.
Start a proxy server, then add `proxy-url` to `~/.kube/config` (note HTTP):

.. code-block: yaml

    clusters:
    - name: …
      cluster:
        server: …
        proxy-url: http://127.0.0.1:8080

Or via the CLI:

    kubectl config get-contexts
    cluster=k3d-k3s-default
    kubectl config set-cluster $cluster --proxy-url=http://127.0.0.1:8080
    kubectl config set-cluster $cluster --proxy-url=  # unset

Written by AI (then adapted to the usage protocol). The prompt is:
Write a simplistic proxy server in Python that serves the CONNECT requests only.
Make it asyncio-based. Listen on 127.0.0.1:8080.
"""
import asyncio
from typing import NamedTuple


class ProxiedRequest(NamedTuple):
    client_host: str
    client_port: int
    server_host: str
    server_port: int


class ProxyServer:
    requests: list[ProxiedRequest]
    host: str
    port: int | None

    def __init__(
            self,
            *,
            host: str = '127.0.0.1',
            port: int | None = None,
    ) -> None:
        super().__init__()
        self.requests = []
        self.host = host
        self.port = port

    @property
    def url(self) -> str:
        return f"http://{self.host}:{self.port}"

    async def run(self) -> None:
        server = await asyncio.start_server(self.handle_client, self.host, self.port)
        addr = server.sockets[0].getsockname()
        self.host = addr[0]
        self.port = addr[1]

        print(f"[*] CONNECT Proxy Server listening on {addr[0]}:{addr[1]}")
        print(f"[*] Press Ctrl+C to stop or cancel programmatically.")

        async with server:
            await server.serve_forever()

    async def handle_client(self, reader, writer):
        """Handle incoming client connection."""
        addr = writer.get_extra_info('peername')
        print(f"[+] Connection from {addr}")

        try:
            # Read the CONNECT request
            request_line = await reader.readline()
            request = request_line.decode('utf-8').strip()

            if not request.startswith('CONNECT'):
                print(f"[-] Non-CONNECT request from {addr}: {request}")
                writer.write(b"HTTP/1.1 405 Method Not Allowed\r\n\r\n")
                await writer.drain()
                writer.close()
                await writer.wait_closed()
                return

            # Parse CONNECT request: CONNECT host:port HTTP/1.1
            parts = request.split()
            if len(parts) < 2:
                print(f"[-] Malformed CONNECT request from {addr}")
                writer.write(b"HTTP/1.1 400 Bad Request\r\n\r\n")
                await writer.drain()
                writer.close()
                await writer.wait_closed()
                return

            target = parts[1]  # host:port

            # Read and discard headers
            while True:
                header = await reader.readline()
                if header == b'\r\n' or header == b'\n' or header == b'':
                    break

            # Parse target host and port
            if ':' in target:
                host, port_str = target.rsplit(':', 1)
                try:
                    port = int(port_str)
                except ValueError:
                    port = 443
            else:
                host = target
                port = 443

            print(f"[*] Connecting to {host}:{port}")

            # Connect to the target server
            try:
                target_reader, target_writer = await asyncio.open_connection(host, port)
            except Exception as e:
                print(f"[-] Failed to connect to {host}:{port} - {e}")
                writer.write(b"HTTP/1.1 502 Bad Gateway\r\n\r\n")
                await writer.drain()
                writer.close()
                await writer.wait_closed()
                return

            # Send success response to client
            writer.write(b"HTTP/1.1 200 Connection Established\r\n\r\n")
            await writer.drain()

            print(f"[+] Tunnel established: {addr} <-> {host}:{port}")
            self.requests.append(ProxiedRequest(addr[0], addr[1], host, port))

            # Run both directions concurrently
            await asyncio.gather(
                self.forward(reader, target_writer, "client->target"),
                self.forward(target_reader, writer, "target->client"),
                return_exceptions=True
            )

            print(f"[-] Tunnel closed: {addr} <-> {host}:{port}")

        except Exception as e:
            print(f"[-] Error handling client {addr}: {e}")
        finally:
            writer.close()
            await writer.wait_closed()

    async def forward(self, src, dst, direction):
        """Bidirectional data forwarding"""
        try:
            while True:
                data = await src.read(8192)
                if not data:
                    break
                dst.write(data)
                await dst.drain()
        except Exception as e:
            print(f"[!] Error in {direction}: {e}")
            raise
        finally:
            dst.close()
            await dst.wait_closed()


if __name__ == '__main__':
    asyncio.run(ProxyServer(port=8080).run())



import asyncio
import contextlib
from unittest.mock import MagicMock

import jsonpatch
import looptime
import pytest

import kopf
from kopf._cogs.aiokits.aiotoggles import ToggleSet
from kopf._cogs.structs.bodies import RawBody
from kopf._cogs.structs.ephemera import Memo
from kopf._core.engines.daemons import daemon_killer
from kopf._core.engines.indexing import OperatorIndexers
from kopf._core.reactor.processing import process_resource_event


class DaemonDummy:

    def __init__(self):
        super().__init__()
        self.mock = MagicMock()

    async def wait_for_daemon_done(self) -> None:
        if not self.mock.called:
            return
        stopped = self.mock.call_args.kwargs['stopped']
        await stopped.wait()
        while stopped.reason is None or not stopped.reason & stopped.reason.DONE:
            await asyncio.sleep(0)  # give control back to asyncio event loop


@pytest.fixture()
async def dummy(simulate_cycle):
    dummy = DaemonDummy()
    yield dummy

    # Cancel the background tasks, if any.
    event_object = {'metadata': {'deletionTimestamp': '...'}}
    with looptime.enabled(strict=True):
        await simulate_cycle(event_object)
        await dummy.wait_for_daemon_done()


@pytest.fixture()
def simulate_cycle(k8s_mocked, registry, settings, resource, memories, mocker):
    """
    Simulate K8s behaviour locally in memory (some meaningful approximation).
    """

    def _merge_dicts(src, dst):
        for key, val in src.items():
            if isinstance(val, dict) and key in dst:
                _merge_dicts(src[key], dst[key])
            else:
                dst[key] = val

    async def _simulate_cycle(
            event_object: RawBody,
            *,
            stream_pressure: asyncio.Event | None = None,
            operator_paused: ToggleSet | None = None,
    ) -> None:
        mocker.resetall()

        await process_resource_event(
            lifecycle=kopf.lifecycles.all_at_once,
            registry=registry,
            settings=settings,
            resource=resource,
            memories=memories,
            memobase=Memo(),
            indexers=OperatorIndexers(),
            raw_event={'type': 'irrelevant', 'object': event_object},
            event_queue=asyncio.Queue(),
            stream_pressure=stream_pressure,
            operator_paused=operator_paused,
            no_throttling=True,
        )

        # Do the same as k8s does: merge the patches into the object.
        for call in k8s_mocked.patch.call_args_list:
            payload = call.kwargs['payload']
            headers = call.kwargs['headers']
            if headers.get('Content-Type') == 'application/merge-patch+json':
                _merge_dicts(payload, event_object)
            if headers.get('Content-Type') == 'application/json-patch+json':
                # For tests, we do not care about resourceVersion checks.
                # Even the json-patch application is optional, but we do it nevertheless.
                payload = [item for item in payload if item['op'] != 'test']
                jsonpatch.JsonPatch(payload).apply(event_object, in_place=True)

    return _simulate_cycle


@pytest.fixture()
async def operator_paused():
    return ToggleSet(any)


@pytest.fixture()
async def conflicts_found(operator_paused: ToggleSet):
    return await operator_paused.make_toggle(name="conflicts_found fixture")


@pytest.fixture()
async def background_daemon_killer(settings, memories, operator_paused):
    """
    Run the daemon killer in the background.
    """
    task = asyncio.create_task(daemon_killer(
        settings=settings, memories=memories, operator_paused=operator_paused))
    yield task

    with contextlib.suppress(asyncio.CancelledError):
        task.cancel()
        await task



import logging

import pytest

from kopf._core.actions.loggers import LocalObjectLogger, ObjectLogger

OBJ1 = {'apiVersion': 'group1/version1', 'kind': 'Kind1',
        'metadata': {'uid': 'uid1', 'name': 'name1', 'namespace': 'ns1'}}
REF1 = {'apiVersion': 'group1/version1', 'kind': 'Kind1',
        'uid': 'uid1', 'name': 'name1', 'namespace': 'ns1'}


@pytest.mark.parametrize('logfn, event_type', [
    ['info', "Normal"],
    ['warning', "Warning"],
    ['error', "Error"],
    ['critical', "Fatal"],
])
async def test_posting_normal_levels(settings, caplog, logstream, logfn, event_type,
                                     event_queue, event_queue_loop):

    logger = ObjectLogger(body=OBJ1, settings=settings)
    logger_fn = getattr(logger, logfn)

    settings.posting.loggers = True
    logger_fn("hello %s", "world")

    assert event_queue.qsize() == 1
    event1 = event_queue.get_nowait()
    assert event1.ref == REF1
    assert event1.type == event_type
    assert event1.reason == "Logging"
    assert event1.message == "hello world"
    assert caplog.messages == ["hello world"]


@pytest.mark.parametrize('logfn, event_type, min_levelno', [
    ['debug', "Debug", logging.DEBUG],
    ['info', "Normal", logging.INFO],
    ['warning', "Warning", logging.WARNING],
    ['error', "Error", logging.ERROR],
    ['critical', "Fatal", logging.CRITICAL],
])
async def test_posting_above_config(settings, caplog, logstream, logfn, event_type, min_levelno,
                                    event_queue, event_queue_loop, mocker):
    logger = ObjectLogger(body=OBJ1, settings=settings)
    logger_fn = getattr(logger, logfn)

    settings.posting.loggers = True
    settings.posting.level = min_levelno
    logger_fn("hello %s", "world")
    settings.posting.level = min_levelno + 1
    logger_fn("must not be posted")

    assert event_queue.qsize() == 1
    event1 = event_queue.get_nowait()
    assert event1.ref == REF1
    assert event1.type == event_type
    assert event1.reason == "Logging"
    assert event1.message == "hello world"
    assert caplog.messages == ["hello world", "must not be posted"]


@pytest.mark.parametrize('logfn', [
    'debug',
])
async def test_skipping_hidden_levels(settings, caplog, logstream, logfn,
                                      event_queue, event_queue_loop):

    logger = ObjectLogger(body=OBJ1, settings=settings)
    logger_fn = getattr(logger, logfn)

    settings.posting.loggers = True
    logger_fn("hello %s", "world")
    logger.info("must be here")

    assert event_queue.qsize() == 1  # not 2!
    assert caplog.messages == ["hello world", "must be here"]


@pytest.mark.parametrize('logfn', [
    'debug',
    'info',
    'warning',
    'error',
    'critical',
])
async def test_skipping_below_config(settings, caplog, logstream, logfn,
                                     event_queue, event_queue_loop, mocker):

    logger = ObjectLogger(body=OBJ1, settings=settings)
    logger_fn = getattr(logger, logfn)

    settings.posting.loggers = True
    settings.posting.level = 666
    logger_fn("hello %s", "world")
    settings.posting.level = 0
    logger.info("must be here")

    assert event_queue.qsize() == 1  # not 2!
    assert caplog.messages == ["hello world", "must be here"]


@pytest.mark.parametrize('logfn', [
    'debug',
    'info',
    'warning',
    'error',
    'critical',
])
@pytest.mark.parametrize('flag', [True, False])
async def test_skipping_when_disabled(settings, caplog, logstream, logfn,
                                      event_queue, event_queue_loop, flag):

    logger = LocalObjectLogger(body=OBJ1, settings=settings)
    logger_fn = getattr(logger, logfn)

    settings.posting.loggers = flag
    settings.posting.enabled = False
    settings.posting.level = 0
    logger_fn("hello %s", "world")

    assert event_queue.qsize() == 0
    assert caplog.messages == ["hello world"]


@pytest.mark.parametrize('logfn', [
    'debug',
    'info',
    'warning',
    'error',
    'critical',
])
@pytest.mark.parametrize('flag', [True, False])
async def test_skipping_when_local_with_all_levels(settings, caplog, logstream, logfn,
                                                   event_queue, event_queue_loop, flag):

    logger = LocalObjectLogger(body=OBJ1, settings=settings)
    logger_fn = getattr(logger, logfn)

    settings.posting.loggers = flag
    logger_fn("hello %s", "world")

    assert event_queue.qsize() == 0
    assert caplog.messages == ["hello world"]



import pytest

from kopf._cogs.structs.credentials import LoginError
from kopf._core.intents.piggybacking import PRIORITY_OF_CLIENT, login_via_client


@pytest.fixture()
def _mock_client(kubernetes, mocker):
    """Mock kubernetes config loading to succeed (in-cluster) and return a known Configuration."""
    mocker.patch.object(kubernetes.config, 'load_incluster_config')
    mocker.patch.object(kubernetes.config, 'load_kube_config')

    config = kubernetes.client.Configuration()
    config.host = 'https://localhost:443'
    config.ssl_ca_cert = '/path/to/ca.crt'
    config.verify_ssl = True
    config.username = 'admin'
    config.password = 'secret'
    config.cert_file = '/path/to/cert.pem'
    config.key_file = '/path/to/key.pem'
    config.api_key = {'authorization': 'Bearer some-token'}
    config.api_key_prefix = {}

    mocker.patch.object(kubernetes.client.Configuration, 'get_default_copy', return_value=config)

    return config


@pytest.mark.usefixtures('no_kubernetes')
def test_returns_none_when_library_is_absent(settings, logger):
    credentials = login_via_client(logger=logger, settings=settings)
    assert credentials is None


def test_in_cluster_config_succeeds(kubernetes, settings, logger, _mock_client):
    credentials = login_via_client(logger=logger, settings=settings)
    assert credentials is not None
    assert kubernetes.config.load_incluster_config.call_count == 1
    assert not kubernetes.config.load_kube_config.called


def test_kubeconfig_fallback_when_incluster_fails(kubernetes, settings, logger, _mock_client):
    kubernetes.config.load_incluster_config.side_effect = kubernetes.config.ConfigException()
    credentials = login_via_client(logger=logger, settings=settings)
    assert credentials is not None
    assert kubernetes.config.load_incluster_config.call_count == 1
    assert kubernetes.config.load_kube_config.call_count == 1


def test_login_error_when_both_configs_fail(kubernetes, settings, logger, _mock_client):
    kubernetes.config.load_incluster_config.side_effect = kubernetes.config.ConfigException()
    kubernetes.config.load_kube_config.side_effect = kubernetes.config.ConfigException()
    with pytest.raises(LoginError, match="Cannot authenticate the client library"):
        login_via_client(logger=logger, settings=settings)


def test_full_credential_extraction(settings, logger, _mock_client):
    credentials = login_via_client(logger=logger, settings=settings)
    assert credentials is not None
    assert credentials.server == 'https://localhost:443'
    assert credentials.ca_path == '/path/to/ca.crt'
    assert credentials.insecure is False  # not verify_ssl (True)
    assert credentials.username == 'admin'
    assert credentials.password == 'secret'
    assert credentials.scheme == 'Bearer'
    assert credentials.token == 'some-token'
    assert credentials.certificate_path == '/path/to/cert.pem'
    assert credentials.private_key_path == '/path/to/key.pem'
    assert credentials.priority == PRIORITY_OF_CLIENT
    assert credentials.default_namespace is None
    assert credentials.proxy_url is None


@pytest.mark.parametrize('header, expected_scheme, expected_token', [
    (None, None, None),
    ('tkn', None, 'tkn'),
    ('Bearer tkn', 'Bearer', 'tkn'),
], ids=['no-header', 'bare-token', 'scheme-and-token'])
def test_token_parsing(settings, logger, _mock_client, header, expected_scheme, expected_token):
    _mock_client.api_key = {'authorization': header} if header else {}
    credentials = login_via_client(logger=logger, settings=settings)
    assert credentials is not None
    assert credentials.scheme == expected_scheme
    assert credentials.token == expected_token


def test_empty_username_and_password_become_none(settings, logger, _mock_client):
    _mock_client.username = ''
    _mock_client.password = ''
    credentials = login_via_client(logger=logger, settings=settings)
    assert credentials is not None
    assert credentials.username is None
    assert credentials.password is None


def test_trust_env_is_propagated_from_settings(settings, logger, _mock_client):
    settings.networking.trust_env = True
    credentials = login_via_client(logger=logger, settings=settings)
    assert credentials is not None
    assert credentials.trust_env is True



import datetime
import pathlib
import ssl

import aiohttp
import pytest

from kopf._cogs.structs.credentials import AiohttpSession, ConnectionInfo, VaultKey


def test_key_as_string():
    key = VaultKey('some-key')
    assert isinstance(key, str)
    assert key == 'some-key'


def test_creation_with_minimal_fields():
    info = ConnectionInfo(
        server='https://localhost',
    )
    assert info.server == 'https://localhost'
    assert info.ca_path is None
    assert info.ca_data is None
    assert info.insecure is None
    assert info.username is None
    assert info.password is None
    assert info.scheme is None
    assert info.token is None
    assert info.certificate_path is None
    assert info.certificate_data is None
    assert info.private_key_path is None
    assert info.private_key_data is None
    assert info.default_namespace is None
    assert info.proxy_url is None
    assert info.priority == 0
    assert info.expiration is None


def test_creation_with_regular_fields():
    info = ConnectionInfo(
        server='https://localhost',
        insecure=True,
        username='username',
        password='password',
        scheme='scheme',
        token='token',
        default_namespace='default',
        proxy_url='https://username:password@localhost',
        priority=123,
        expiration=datetime.datetime.max,
    )
    assert info.server == 'https://localhost'
    assert info.insecure is True
    assert info.username == 'username'
    assert info.password == 'password'
    assert info.scheme == 'scheme'
    assert info.token == 'token'
    assert info.default_namespace == 'default'
    assert info.priority == 123
    assert info.proxy_url == 'https://username:password@localhost'
    assert info.expiration == datetime.datetime.max


def test_creation_with_ssl_data_without_files():
    info = ConnectionInfo(
        server='https://localhost',
        ca_data=b'ca_data',
        certificate_data=b'cert_data',
        private_key_data=b'pkey_data',
    )
    assert info.server == 'https://localhost'
    assert info.ca_path is None
    assert info.ca_data == b'ca_data'
    assert info.certificate_path is None
    assert info.certificate_data == b'cert_data'
    assert info.private_key_path is None
    assert info.private_key_data == b'pkey_data'


def test_creation_with_ssl_path_without_data():
    info = ConnectionInfo(
        server='https://localhost',
        ca_path='/ca/path',
        certificate_path='/cert/path',
        private_key_path='/pkey/path',
    )
    assert info.server == 'https://localhost'
    assert info.ca_path == '/ca/path'
    assert info.ca_data is None
    assert info.certificate_path == '/cert/path'
    assert info.certificate_data is None
    assert info.private_key_path == '/pkey/path'
    assert info.private_key_data is None


def test_conflicting_ca_data_and_path():
    with pytest.raises(ValueError, match="Both CA path & data"):
        ConnectionInfo(server='', ca_path='/path', ca_data=b'data')


def test_conflicting_certificate_data_and_path():
    with pytest.raises(ValueError, match="Both certificate path & data"):
        ConnectionInfo(server='', certificate_path='/path', certificate_data=b'data')


def test_conflicting_private_key_data_and_path():
    with pytest.raises(ValueError, match="Both private key path & data"):
        ConnectionInfo(server='', private_key_path='/path', private_key_data=b'data')


def test_connection_info_as_aiohttp_basic_auth():
    info = ConnectionInfo(
        server='https://localhost',
        username='username',
        password='password',
    )
    assert info.as_aiohttp_basic_auth() == aiohttp.BasicAuth('username', 'password')
    assert info.as_http_headers() == {}


def test_connection_info_as_http_headers():
    info = ConnectionInfo(
        server='https://localhost',
        scheme='Bearer',
        token='xyz'
    )
    assert info.as_aiohttp_basic_auth() is None
    assert info.as_http_headers() == {'Authorization': 'Bearer xyz'}


def test_connection_info_as_ssl_context_when_insecure():
    info = ConnectionInfo(
        server='https://localhost',
        insecure=True,
    )
    ssl_context = info.as_ssl_context()
    ca = ssl_context.get_ca_certs()
    assert ssl_context.verify_mode == ssl.CERT_NONE
    assert ssl_context.check_hostname is False
    assert ca  # at least some default CAs must be loaded, but we do not know which ones.


def test_connection_info_as_ssl_context_when_defined():
    info = ConnectionInfo(
        server='https://localhost',
        ca_path=pathlib.Path(__file__).parent / 'fixtures/ca.pem',
        certificate_path=pathlib.Path(__file__).parent / 'fixtures/cert.pem',
        private_key_path=pathlib.Path(__file__).parent / 'fixtures/pkey.pem',
    )
    ssl_context = info.as_ssl_context()
    ca = ssl_context.get_ca_certs()
    assert ssl_context.verify_mode == ssl.CERT_REQUIRED
    assert ssl_context.check_hostname is True
    assert ca[0]['issuer'] == ((('commonName', 'minikubeCA'),),)
    assert ca[0]['subject'] == ((('commonName', 'minikubeCA'),),)
    assert ca[0]['notAfter'] == 'May 19 09:18:36 2029 GMT'
    assert ca[0]['notBefore'] == 'May 21 09:18:36 2019 GMT'


async def test_creation_of_aiohttp_session():
    aiohttp_session = aiohttp.ClientSession()
    info = AiohttpSession(
        server='https://localhost',
        default_namespace='default',
        priority=123,
        expiration=datetime.datetime.max,
        aiohttp_session=aiohttp_session,
    )
    assert info.server == 'https://localhost'
    assert info.default_namespace == 'default'
    assert info.priority == 123
    assert info.expiration == datetime.datetime.max
    assert info.aiohttp_session is aiohttp_session
    await aiohttp_session.close()



import pytest

from kopf._cogs.clients.errors import APIError
from kopf._cogs.clients.fetching import list_objs


async def test_listing_works(kmock, settings, logger, resource, namespace):
    kmock[resource, kmock.namespace(namespace)] << {'items': [{}, {}]}
    items, resource_version = await list_objs(
        logger=logger,
        settings=settings,
        resource=resource,
        namespace=namespace,
    )
    assert items == [{}, {}]
    assert len(kmock['list']) == 1


# Note: 401 is wrapped into a LoginError and is tested elsewhere.
@pytest.mark.parametrize('status', [400, 403, 500, 666])
async def test_raises_direct_api_errors(
        kmock, settings, logger, status, resource, namespace,
        cluster_resource, namespaced_resource):
    kmock[cluster_resource, kmock.namespace(None)] << status
    kmock[namespaced_resource, kmock.namespace('ns')] << status

    with pytest.raises(APIError) as e:
        await list_objs(
            logger=logger,
            settings=settings,
            resource=resource,
            namespace=namespace,
        )
    assert e.value.status == status



import functools
from collections.abc import Callable
from contextvars import ContextVar
from typing import Any, TypeVar, cast

import aiohttp

from kopf._cogs.clients import errors
from kopf._cogs.helpers import versions
from kopf._cogs.structs import credentials

# Per-operator storage and exchange point for authentication methods.
# Used by the client wrappers to retrieve the credentials and report the failures.
# Set by `spawn_tasks`, so that every operator's task has the same vault.
vault_var: ContextVar[credentials.Vault] = ContextVar('vault_var')

# A typevar to show that we return a function with the same signature as given.
_F = TypeVar('_F', bound=Callable[..., Any])


def authenticated(fn: _F) -> _F:
    """
    A decorator to inject a pre-authenticated session to a requesting routine.

    If a wrapped function fails on the authentication, this will be reported
    back to the credentials vault, which will trigger the re-authentication
    activity. Meanwhile, the request-performing function will be awaiting
    new credentials, and re-executed once they are available.
    """
    @functools.wraps(fn)
    async def wrapper(*args: Any, **kwargs: Any) -> Any:

        # If a context is explicitly passed, make it a simple call without re-auth.
        # Exceptions are escalated to a caller, which is probably wrapped itself.
        if 'context' in kwargs:
            context = kwargs['context']
            response = await fn(*args, **kwargs)
            if isinstance(response, aiohttp.ClientResponse):
                # Keep track of responses which are using this context.
                context.add_response(response)
            return response

        # Otherwise, attempt the execution with the vault credentials and re-authenticate on 401s.
        vault: credentials.Vault = vault_var.get()
        async for key, info, context in vault.extended(APIContext, 'contexts'):
            try:
                response = await fn(*args, **kwargs, context=context)
                if isinstance(response, aiohttp.ClientResponse):
                    # Keep track of responses which are using this context.
                    context.add_response(response)
                return response
            except (errors.APIUnauthorizedError, errors.APISessionClosed) as e:
                await vault.invalidate(key, info, exc=e)

        # Normally, either `vault.extended()` or `vault.invalidate()` raise the login errors.
        # The for-cycle can only end if the yielded credentials are not invalidated before trying
        # the next ones -- but this case exits by `return` or by other (non-401) errors.
        raise RuntimeError("Reached an impossible state: the end of the authentication cycle.")

    return cast(_F, wrapper)


class APIContext:
    """
    A container for an aiohttp session and the caches of the environment info.

    The container is constructed only once for every :class:`KubeContext`,
    and then cached for later re-use (see :meth:`Vault.extended`).

    We assume that the whole operator runs in the same event loop, so there is
    no need to split the sessions for multiple loops. Synchronous handlers are
    threaded with other event loops per thread, but no operator's requests are
    performed inside of those threads: everything is in the main thread/loop.
    """

    # The main contained object used by the API methods.
    session: aiohttp.ClientSession

    # Contextual information for URL building.
    server: str
    default_namespace: str | None

    # List of open responses.
    responses: list[aiohttp.ClientResponse]

    def __init__(
            self,
            info: credentials.KubeContext,
    ) -> None:
        super().__init__()

        # Generic aiohttp session based on the constructed credentials.
        match info:
            case credentials.ConnectionInfo():
                self.session = aiohttp.ClientSession(
                    connector=aiohttp.TCPConnector(
                        limit=0,
                        ssl=info.as_ssl_context(),
                    ),
                    headers=info.as_http_headers(),
                    auth=info.as_aiohttp_basic_auth(),
                    proxy=info.proxy_url,
                    trust_env=info.trust_env,
                )
            case credentials.AiohttpSession():
                self.session = info.aiohttp_session
            case _:
                raise TypeError(f"Unsupported credentials type: {info!r}")

        # It is a good practice to self-identify a bit, even with a user-provided session.
        if self.session.headers.get('User-Agent') is None:
            self.session.headers['User-Agent'] = f'kopf/{versions.version or "unknown"}'

        # Add the extra payload information. We avoid overriding the constructor.
        self.server = info.server
        self.default_namespace = info.default_namespace

        self.responses = []

    def flush_closed_responses(self) -> None:
        # There's no point keeping references to already closed responses.
        self.responses[:] = [_response for _response in self.responses if not _response.closed]

    def add_response(self, response: aiohttp.ClientResponse) -> None:
        # Keep track of responses so they can be closed later when the session
        # is closed.
        self.flush_closed_responses()
        if not response.closed:
            self.responses.append(response)

    def close_open_responses(self) -> None:
        # Close all responses that are still open and are using this session.
        for response in self.responses:
            if not response.closed:
                response.close()
        self.responses.clear()

    async def close(self) -> None:
        # Close all open responses that use this session before closing the session itself.
        self.close_open_responses()

        # Closing is triggered by `Vault._flush_caches()` -- forward it to the actual session.
        await self.session.close()



import asyncio

import pytest

from kopf._cogs.aiokits.aiobindings import condition_chain


async def test_no_triggering(looptime):
    source = asyncio.Condition()
    target = asyncio.Condition()
    task = asyncio.create_task(condition_chain(source, target))
    try:
        with pytest.raises(asyncio.TimeoutError):
            async with target:
                await asyncio.wait_for(target.wait(), timeout=1.23)
        assert looptime == 1.23
    finally:
        task.cancel()
        await asyncio.wait([task])


async def test_triggering(looptime):
    source = asyncio.Condition()
    target = asyncio.Condition()
    task = asyncio.create_task(condition_chain(source, target))
    try:

        async def delayed_trigger():
            async with source:
                source.notify_all()

        loop = asyncio.get_running_loop()
        loop.call_later(1.23, asyncio.create_task, delayed_trigger())

        async with target:
            await target.wait()

        assert looptime == 1.23

    finally:
        task.cancel()
        await asyncio.wait([task])



from typing import Any

import kopf


@kopf.on.create('kopfexamples')
def create_fn(spec: kopf.Spec, **_: Any) -> None:

    for item in spec.get('items', []):

        @kopf.subhandler(id=item)
        async def create_item_fn(item: str = item, **_: Any) -> None:
            print(f"=== Handling creation for {item}. ===")



import json

import pytest

from kopf._cogs.configs.conventions import StorageKeyFormingConvention
from kopf._cogs.configs.diffbase import AnnotationsDiffBaseStorage
from kopf._cogs.configs.progress import AnnotationsProgressStorage, \
                                        ProgressRecord, SmartProgressStorage
from kopf._cogs.structs.bodies import Body
from kopf._cogs.structs.ids import HandlerId
from kopf._cogs.structs.patches import Patch

ANNOTATIONS_POPULATING_STORAGES = [
    AnnotationsProgressStorage,
    SmartProgressStorage,
]
STORAGE_KEY_FORMING_CLASSES = [
    StorageKeyFormingConvention,
    AnnotationsProgressStorage,
    AnnotationsDiffBaseStorage,
]

CONTENT_DATA = ProgressRecord(
    started='2020-01-01T00:00:00',
    stopped='2020-12-31T23:59:59',
    delayed='3000-01-01T00:00:00',
    retries=0,
    success=False,
    failure=False,
    message=None,
)

CONTENT_JSON = json.dumps(CONTENT_DATA, separators=(',', ':'))

COMMON_KEYS = [
    # For character replacements (only those that happen in our own ids, not all of them).
    ['my-operator.example.com', 'a_b.c-d/e', 'my-operator.example.com/a_b.c-d.e'],
]

V1_KEYS = [
    # For V1 length cutting. Hint: the prefix length is 23, the remaining space is 63 - 23 - 1 = 39.
    # The suffix itself (if appended) takes 9, so it is 30 left. The same math for no prefix.
    ['my-operator.example.com', 'x', 'my-operator.example.com/x'],
    ['my-operator.example.com', 'x' * 39, 'my-operator.example.com/' + 'x' * 39],
    ['my-operator.example.com', 'x' * 40, 'my-operator.example.com/' + 'x' * 30 + 'xx-tEokcg'],
    ['my-operator.example.com', 'y' * 40, 'my-operator.example.com/' + 'y' * 30 + 'yy-VZlvhw'],
    ['my-operator.example.com', 'z' * 40, 'my-operator.example.com/' + 'z' * 30 + 'zz-LlPQyA'],

    # For special chars in base64 encoding ("+" and "/"), which are not compatible with K8s.
    # The numbers are found empirically so that both "/" and "+" are found in the base64'ed digest.
    ['my-operator.example.com', 'fn' * 323, 'my-operator.example.com/' + 'fn' * 15 + 'fn-Az-r.g'],
]

V2_KEYS = [
    # For V2 length cutting: 63 for the name part only, not the whole annotation.
    # The suffix itself (if appended) takes 9, so it is 63-9=54 left.
    ['my-operator.example.com', 'x', 'my-operator.example.com/x'],
    ['my-operator.example.com', 'x' * 63, 'my-operator.example.com/' + 'x' * 63],
    ['my-operator.example.com', 'x' * 64, 'my-operator.example.com/' + 'x' * 54 + 'xx-SItAqA'],
    ['my-operator.example.com', 'y' * 64, 'my-operator.example.com/' + 'y' * 54 + 'yy-0d251g'],
    ['my-operator.example.com', 'z' * 64, 'my-operator.example.com/' + 'z' * 54 + 'zz-E7wvIA'],

    # For special chars in base64 encoding ("+" and "/"), which are not compatible with K8s.
    # The numbers are found empirically so that both "/" and "+" are found in the base64'ed digest.
    ['my-operator.example.com', 'fn' * 323, 'my-operator.example.com/' + 'fn' * 27 + 'fn-Az-r.g'],
]


@pytest.mark.parametrize('cls', STORAGE_KEY_FORMING_CLASSES)
def test_keys_for_all_versions(cls):
    storage = cls(v1=True, prefix='kopf.zalando.org')
    v1_key = storage.make_v1_key('.' * 64)
    v2_key = storage.make_v2_key('.' * 64)
    assert v1_key != v2_key  # prerequisite
    keys = storage.make_keys('.' * 64)
    assert len(list(keys)) == 2
    assert v1_key in keys
    assert v2_key in keys


@pytest.mark.parametrize('cls', STORAGE_KEY_FORMING_CLASSES)
def test_keys_deduplication(cls):
    storage = cls(v1=True, prefix='kopf.zalando.org')
    v1_key = storage.make_v1_key('...')
    v2_key = storage.make_v2_key('...')
    assert v1_key == v2_key  # prerequisite
    keys = storage.make_keys('...')
    assert len(list(keys)) == 1
    assert v1_key in keys
    assert v2_key in keys


@pytest.mark.parametrize('kind, owners, expected', [
    pytest.param('ReplicaSet', [{'kind': 'Deployment'}], 'kopf.dev/xyz-ofDRS', id='DRS'),
    pytest.param('ReplicaSet', [{'kind': 'OtherOwner'}], 'kopf.dev/xyz', id='not-deployment'),
    pytest.param('OtherKind', [{'kind': 'Deployment'}], 'kopf.dev/xyz', id='not-replicaset'),
])
@pytest.mark.parametrize('cls', STORAGE_KEY_FORMING_CLASSES)
def test_keys_of_replicaset_owned_by_deployment(cls, kind, owners, expected):
    storage = cls(v1=True, prefix='kopf.dev')
    body = Body({'kind': kind, 'metadata': {'ownerReferences': owners}})
    keys = storage.make_keys('xyz', body=body)
    assert set(keys) == {expected}


@pytest.mark.parametrize('prefix, provided_key, expected_key', COMMON_KEYS + V1_KEYS)
@pytest.mark.parametrize('cls', STORAGE_KEY_FORMING_CLASSES)
def test_key_hashing_v1(cls, prefix, provided_key, expected_key):
    storage = cls(v1=True, prefix=prefix)
    returned_key = storage.make_v1_key(provided_key)
    assert returned_key == expected_key


@pytest.mark.parametrize('prefix, provided_key, expected_key', COMMON_KEYS + V2_KEYS)
@pytest.mark.parametrize('cls', STORAGE_KEY_FORMING_CLASSES)
def test_key_hashing_v2(cls, prefix, provided_key, expected_key):
    storage = cls(v1=True, prefix=prefix)
    returned_key = storage.make_v2_key(provided_key)
    assert returned_key == expected_key


@pytest.mark.parametrize('prefix, provided_key, expected_key', COMMON_KEYS + V1_KEYS + V2_KEYS)
@pytest.mark.parametrize('cls', ANNOTATIONS_POPULATING_STORAGES)
def test_keys_hashed_on_fetching(cls, prefix, provided_key, expected_key):
    storage = cls(prefix=prefix)
    body = Body({'metadata': {'annotations': {expected_key: CONTENT_JSON}}})
    record = storage.fetch(body=body, key=HandlerId(provided_key))
    assert record is not None
    assert record == CONTENT_DATA


@pytest.mark.parametrize('prefix, provided_key, expected_key', COMMON_KEYS + V1_KEYS + V2_KEYS)
@pytest.mark.parametrize('cls', ANNOTATIONS_POPULATING_STORAGES)
def test_keys_normalized_on_storing(cls, prefix, provided_key, expected_key):
    storage = cls(prefix=prefix)
    patch = Patch()
    body = Body({'metadata': {'annotations': {expected_key: 'null'}}})
    storage.store(body=body, patch=patch, key=HandlerId(provided_key), record=CONTENT_DATA)
    assert expected_key in patch.metadata.annotations


@pytest.mark.parametrize('prefix, provided_key, expected_key', COMMON_KEYS + V1_KEYS + V2_KEYS)
@pytest.mark.parametrize('cls', ANNOTATIONS_POPULATING_STORAGES)
def test_keys_normalized_on_purging(cls, prefix, provided_key, expected_key):
    storage = cls(prefix=prefix)
    patch = Patch()
    body = Body({'metadata': {'annotations': {expected_key: 'null'}}})
    storage.purge(body=body, patch=patch, key=HandlerId(provided_key))
    assert expected_key in patch.metadata.annotations


@pytest.mark.parametrize('prefix, provided_key, expected_key', COMMON_KEYS + V1_KEYS + V2_KEYS)
@pytest.mark.parametrize('cls', ANNOTATIONS_POPULATING_STORAGES)
def test_keys_normalized_on_touching(cls, prefix, provided_key, expected_key):
    storage = cls(prefix=prefix, touch_key=provided_key)
    patch = Patch()
    body = Body({})
    storage.touch(body=body, patch=patch, value='irrelevant')
    assert expected_key in patch.metadata.annotations


@pytest.mark.parametrize('cls', ANNOTATIONS_POPULATING_STORAGES + STORAGE_KEY_FORMING_CLASSES)
def test_warning_on_long_prefix(cls):
    with pytest.warns(UserWarning, match=r"The annotations prefix is too long"):
        cls(v1=True, prefix='x' * (253 - 63))



import copy

import pytest

import kopf

forced_mode = pytest.mark.parametrize('forcedness', [
    pytest.param(dict(forced=True), id='forcedTrue'),
])
non_forced_mode = pytest.mark.parametrize('forcedness', [
    pytest.param(dict(forced=False), id='forcedFalse'),
    pytest.param(dict(), id='forcedAbsent'),
])
any_forced_mode = pytest.mark.parametrize('forcedness', [
    pytest.param(dict(forced=True), id='forcedTrue'),
    pytest.param(dict(forced=False), id='forcedFalse'),
    pytest.param(dict(), id='forcedAbsent'),
])

strict_mode = pytest.mark.parametrize('strictness', [
    pytest.param(dict(strict=True), id='strictTrue'),
])
non_strict_mode = pytest.mark.parametrize('strictness', [
    pytest.param(dict(strict=False), id='strictFalse'),
    pytest.param(dict(), id='strictAbsent'),
])
any_strict_mode = pytest.mark.parametrize('strictness', [
    pytest.param(dict(strict=True), id='strictTrue'),
    pytest.param(dict(strict=False), id='strictFalse'),
    pytest.param(dict(), id='strictAbsent'),
])

obj1_with_names = pytest.mark.parametrize('obj1', [
    pytest.param({'metadata': {'name': 'a'}}, id='regularname'),
    pytest.param({'metadata': {'generateName': 'b'}}, id='generatename'),
    pytest.param({'metadata': {'name': 'c', 'generateName': 'd'}}, id='bothnames'),
])
obj2_with_names = pytest.mark.parametrize('obj2', [
    pytest.param({'metadata': {'name': 'a'}}, id='regularname'),
    pytest.param({'metadata': {'generateName': 'b'}}, id='generatename'),
    pytest.param({'metadata': {'name': 'c', 'generateName': 'd'}}, id='bothnames'),
])
obj1_without_names = pytest.mark.parametrize('obj1', [
    pytest.param({}, id='withoutmeta'),
    pytest.param({'metadata': {}}, id='withmeta'),
])
obj2_without_names = pytest.mark.parametrize('obj2', [
    pytest.param({}, id='withoutmeta'),
    pytest.param({'metadata': {}}, id='withmeta'),
])


# In the NON-FORCED mode, the EXISTING names are preserved.
# The strictness is not involved due to this (no new names added).
@obj1_with_names
@any_strict_mode
@non_forced_mode
def test_preserved_name_of_dict(forcedness, strictness, obj1):
    obj1 = copy.deepcopy(obj1)
    kopf.harmonize_naming(obj1, name='provided-name', **forcedness, **strictness)
    assert obj1['metadata'].get('name') != 'provided-name'
    assert obj1['metadata'].get('generateName') != 'provided-name'


@obj2_with_names
@obj1_with_names
@any_strict_mode
@non_forced_mode
def test_preserved_names_of_dicts(forcedness, strictness, multicls, obj1, obj2):
    obj1, obj2 = copy.deepcopy(obj1), copy.deepcopy(obj2)
    objs = multicls([obj1, obj2])
    kopf.harmonize_naming(objs, name='provided-name', **forcedness, **strictness)
    assert obj1['metadata'].get('name') != 'provided-name'
    assert obj2['metadata'].get('name') != 'provided-name'
    assert obj1['metadata'].get('generateName') != 'provided-name'
    assert obj2['metadata'].get('generateName') != 'provided-name'


@obj1_with_names
@any_strict_mode
@non_forced_mode
def test_preserved_names_of_pykube_object(forcedness, strictness, pykube_object, obj1):
    pykube_object.obj = copy.deepcopy(obj1)
    kopf.harmonize_naming(pykube_object, name='provided-name', **forcedness, **strictness)
    assert pykube_object.obj['metadata'].get('name') != 'provided-name'
    assert pykube_object.obj['metadata'].get('generateName') != 'provided-name'


@obj1_with_names
@any_strict_mode
@non_forced_mode
def test_preserved_names_of_kubernetes_model(forcedness, strictness, kubernetes_model, obj1):
    kubernetes_model.metadata.name = obj1.get('metadata', {}).get('name')
    kubernetes_model.metadata.generate_name = obj1.get('metadata', {}).get('generateName')
    kopf.harmonize_naming(kubernetes_model, name='provided-name', **forcedness, **strictness)
    assert kubernetes_model.metadata.name != 'provided-name'
    assert kubernetes_model.metadata.generate_name != 'provided-name'


@obj1_with_names
@any_strict_mode
@non_forced_mode
def test_preserved_names_of_kubernetes_asyncio_model(forcedness, strictness, kubernetes_asyncio_model, obj1):
    kubernetes_asyncio_model.metadata.name = obj1.get('metadata', {}).get('name')
    kubernetes_asyncio_model.metadata.generate_name = obj1.get('metadata', {}).get('generateName')
    kopf.harmonize_naming(kubernetes_asyncio_model, name='provided-name', **forcedness, **strictness)
    assert kubernetes_asyncio_model.metadata.name != 'provided-name'
    assert kubernetes_asyncio_model.metadata.generate_name != 'provided-name'


# In the FORCED mode, the EXISTING names are overwritten.
# It only depends which of the names -- regular or generated -- is left.
@obj1_with_names
@strict_mode
@forced_mode
def test_overwriting_of_strict_name_of_dict(forcedness, strictness, obj1):
    obj1 = copy.deepcopy(obj1)
    kopf.harmonize_naming(obj1, name='provided-name', **forcedness, **strictness)
    assert 'name' in obj1['metadata']
    assert 'generateName' not in obj1['metadata']
    assert obj1['metadata']['name'] == 'provided-name'


@obj2_with_names
@obj1_with_names
@strict_mode
@forced_mode
def test_overwriting_of_strict_names_of_dicts(forcedness, strictness, multicls, obj1, obj2):
    obj1, obj2 = copy.deepcopy(obj1), copy.deepcopy(obj2)
    objs = multicls([obj1, obj2])
    kopf.harmonize_naming(objs, name='provided-name', **forcedness, **strictness)
    assert 'name' in obj1['metadata']
    assert 'name' in obj2['metadata']
    assert 'generateName' not in obj1['metadata']
    assert 'generateName' not in obj2['metadata']
    assert obj2['metadata']['name'] == 'provided-name'
    assert obj1['metadata']['name'] == 'provided-name'


@obj1_with_names
@strict_mode
@forced_mode
def test_overwriting_of_strict_name_of_pykube_object(forcedness, strictness, pykube_object, obj1):
    pykube_object.obj = copy.deepcopy(obj1)
    kopf.harmonize_naming(pykube_object, name='provided-name', **forcedness, **strictness)
    assert pykube_object.obj['metadata'].get('name') == 'provided-name'
    assert pykube_object.obj['metadata'].get('generateName') is None


@obj1_with_names
@strict_mode
@forced_mode
def test_overwriting_of_strict_name_of_kubernetes_model(forcedness, strictness, kubernetes_model, obj1):
    kubernetes_model.metadata.name = obj1.get('metadata', {}).get('name')
    kubernetes_model.metadata.generate_name = obj1.get('metadata', {}).get('generateName')
    kopf.harmonize_naming(kubernetes_model, name='provided-name', **forcedness, **strictness)
    assert kubernetes_model.metadata.name == 'provided-name'
    assert kubernetes_model.metadata.generate_name is None


@obj1_with_names
@strict_mode
@forced_mode
def test_overwriting_of_strict_name_of_kubernetes_asyncio_model(forcedness, strictness, kubernetes_asyncio_model, obj1):
    kubernetes_asyncio_model.metadata.name = obj1.get('metadata', {}).get('name')
    kubernetes_asyncio_model.metadata.generate_name = obj1.get('metadata', {}).get('generateName')
    kopf.harmonize_naming(kubernetes_asyncio_model, name='provided-name', **forcedness, **strictness)
    assert kubernetes_asyncio_model.metadata.name == 'provided-name'
    assert kubernetes_asyncio_model.metadata.generate_name is None


@obj1_with_names
@non_strict_mode
@forced_mode
def test_overwriting_of_relaxed_name_of_dict(forcedness, strictness, obj1):
    obj1 = copy.deepcopy(obj1)
    kopf.harmonize_naming(obj1, name='provided-name', **forcedness, **strictness)
    assert 'name' not in obj1['metadata']
    assert 'generateName' in obj1['metadata']
    assert obj1['metadata']['generateName'] == 'provided-name-'


@obj2_with_names
@obj1_with_names
@non_strict_mode
@forced_mode
def test_overwriting_of_relaxed_names_of_dicts(forcedness, strictness, multicls, obj1, obj2):
    obj1, obj2 = copy.deepcopy(obj1), copy.deepcopy(obj2)
    objs = multicls([obj1, obj2])
    kopf.harmonize_naming(objs, name='provided-name', **forcedness, **strictness)
    assert 'name' not in obj1['metadata']
    assert 'name' not in obj2['metadata']
    assert 'generateName' in obj1['metadata']
    assert 'generateName' in obj2['metadata']
    assert obj1['metadata']['generateName'] == 'provided-name-'
    assert obj2['metadata']['generateName'] == 'provided-name-'


@obj1_with_names
@non_strict_mode
@forced_mode
def test_overwriting_of_relaxed_name_of_pykube_object(forcedness, strictness, pykube_object, obj1):
    pykube_object.obj = copy.deepcopy(obj1)
    kopf.harmonize_naming(pykube_object, name='provided-name', **forcedness, **strictness)
    assert pykube_object.obj['metadata'].get('name') is None
    assert pykube_object.obj['metadata'].get('generateName') == 'provided-name-'


@obj1_with_names
@non_strict_mode
@forced_mode
def test_overwriting_of_relaxed_name_of_kubernetes_model(forcedness, strictness, kubernetes_model, obj1):
    kubernetes_model.metadata.name = obj1.get('metadata', {}).get('name')
    kubernetes_model.metadata.generate_name = obj1.get('metadata', {}).get('generateName')
    kopf.harmonize_naming(kubernetes_model, name='provided-name', **forcedness, **strictness)
    assert kubernetes_model.metadata.name is None
    assert kubernetes_model.metadata.generate_name == 'provided-name-'


@obj1_with_names
@non_strict_mode
@forced_mode
def test_overwriting_of_relaxed_name_of_kubernetes_asyncio_model(forcedness, strictness, kubernetes_asyncio_model, obj1):
    kubernetes_asyncio_model.metadata.name = obj1.get('metadata', {}).get('name')
    kubernetes_asyncio_model.metadata.generate_name = obj1.get('metadata', {}).get('generateName')
    kopf.harmonize_naming(kubernetes_asyncio_model, name='provided-name', **forcedness, **strictness)
    assert kubernetes_asyncio_model.metadata.name is None
    assert kubernetes_asyncio_model.metadata.generate_name == 'provided-name-'


# When names are ABSENT, they are added regardless of the forced mode.
# The only varying part is which name is added: regular or generated.
@obj1_without_names
@strict_mode
@any_forced_mode
def test_assignment_of_strict_name_of_dict(forcedness, strictness, obj1):
    obj1 = copy.deepcopy(obj1)
    kopf.harmonize_naming(obj1, name='provided-name', **forcedness, **strictness)
    assert 'name' in obj1['metadata']
    assert 'generateName' not in obj1['metadata']
    assert obj1['metadata']['name'] == 'provided-name'


@obj2_without_names
@obj1_without_names
@strict_mode
@any_forced_mode
def test_assignment_of_strict_names_of_dicts(forcedness, strictness, multicls, obj1, obj2):
    obj1, obj2 = copy.deepcopy(obj1), copy.deepcopy(obj2)
    objs = multicls([obj1, obj2])
    kopf.harmonize_naming(objs, name='provided-name', **forcedness, **strictness)
    assert 'name' in obj1['metadata']
    assert 'name' in obj2['metadata']
    assert 'generateName' not in obj1['metadata']
    assert 'generateName' not in obj2['metadata']
    assert obj1['metadata']['name'] == 'provided-name'
    assert obj2['metadata']['name'] == 'provided-name'


@obj1_without_names
@strict_mode
@any_forced_mode
def test_assignment_of_strict_name_of_pykube_object(forcedness, strictness, pykube_object, obj1):
    pykube_object.obj = copy.deepcopy(obj1)
    kopf.harmonize_naming(pykube_object, name='provided-name', **forcedness, **strictness)
    assert pykube_object.obj['metadata'].get('name') == 'provided-name'
    assert pykube_object.obj['metadata'].get('generateName') is None


@strict_mode
@any_forced_mode
def test_assignment_of_strict_name_of_kubernetes_model(forcedness, strictness, kubernetes_model):
    kubernetes_model.metadata = None
    kopf.harmonize_naming(kubernetes_model, name='provided-name', **forcedness, **strictness)
    assert kubernetes_model.metadata.name == 'provided-name'
    assert kubernetes_model.metadata.generate_name is None


@strict_mode
@any_forced_mode
def test_assignment_of_strict_name_of_kubernetes_asyncio_model(forcedness, strictness, kubernetes_asyncio_model):
    kubernetes_asyncio_model.metadata = None
    kopf.harmonize_naming(kubernetes_asyncio_model, name='provided-name', **forcedness, **strictness)
    assert kubernetes_asyncio_model.metadata.name == 'provided-name'
    assert kubernetes_asyncio_model.metadata.generate_name is None


@obj1_without_names
@non_strict_mode
@any_forced_mode
def test_assignment_of_nonstrict_name_of_dict(forcedness, strictness, obj1):
    obj1 = copy.deepcopy(obj1)
    kopf.harmonize_naming(obj1, name='provided-name', **forcedness, **strictness)
    assert 'name' not in obj1['metadata']
    assert 'generateName' in obj1['metadata']
    assert obj1['metadata']['generateName'] == 'provided-name-'


@obj2_without_names
@obj1_without_names
@non_strict_mode
@any_forced_mode
def test_assignment_of_nonstrict_names_of_dicts(forcedness, strictness, multicls, obj1, obj2):
    obj1, obj2 = copy.deepcopy(obj1), copy.deepcopy(obj2)
    objs = multicls([obj1, obj2])
    kopf.harmonize_naming(objs, name='provided-name', **forcedness, **strictness)
    assert 'name' not in obj1['metadata']
    assert 'name' not in obj2['metadata']
    assert 'generateName' in obj1['metadata']
    assert 'generateName' in obj2['metadata']
    assert obj1['metadata']['generateName'] == 'provided-name-'
    assert obj2['metadata']['generateName'] == 'provided-name-'


@obj1_without_names
@non_strict_mode
@any_forced_mode
def test_assignment_of_nonstrict_name_of_pykube_object(forcedness, strictness, pykube_object, obj1):
    pykube_object.obj = copy.deepcopy(obj1)
    kopf.harmonize_naming(pykube_object, name='provided-name', **forcedness, **strictness)
    assert pykube_object.obj['metadata'].get('name') is None
    assert pykube_object.obj['metadata'].get('generateName') == 'provided-name-'


@non_strict_mode
@any_forced_mode
def test_assignment_of_nonstrict_name_of_kubernetes_model(forcedness, strictness, kubernetes_model):
    kubernetes_model.metadata = None
    kopf.harmonize_naming(kubernetes_model, name='provided-name', **forcedness, **strictness)
    assert kubernetes_model.metadata.name is None
    assert kubernetes_model.metadata.generate_name == 'provided-name-'


@non_strict_mode
@any_forced_mode
def test_assignment_of_nonstrict_name_of_kubernetes_asyncio_model(forcedness, strictness, kubernetes_asyncio_model):
    kubernetes_asyncio_model.metadata = None
    kopf.harmonize_naming(kubernetes_asyncio_model, name='provided-name', **forcedness, **strictness)
    assert kubernetes_asyncio_model.metadata.name is None
    assert kubernetes_asyncio_model.metadata.generate_name == 'provided-name-'



import asyncio


async def condition_chain(
        source: asyncio.Condition,
        target: asyncio.Condition,
) -> None:
    """
    A condition chain is a "clean" hack to attach one condition to another.

    It is a "clean" (not "dirty") hack to wake up the webhook configuration
    managers when either the resources are revised (as seen in the insights),
    or a new client config is yielded from the webhook server.
    """
    async with source:
        while True:
            await source.wait()
            async with target:
                target.notify_all()



from typing import Any

import kopf


@kopf.timer('kopfexamples', idle=5, interval=2)
def every_few_seconds_sync(spec: kopf.Spec, logger: kopf.Logger, **_: Any) -> None:
    logger.info(f"Ping from a sync timer: field={spec['field']!r}")


@kopf.timer('kopfexamples', idle=10, interval=4)
async def every_few_seconds_async(spec: kopf.Spec, logger: kopf.Logger, **_: Any) -> None:
    logger.info(f"Ping from an async timer: field={spec['field']!r}")



from typing import Any

import kopf


@kopf.on.create('kopfexamples')
def create_fn(spec: kopf.Spec, **_: Any) -> Any:
    print(f"And here we are! Creating: {spec}")
    return {'message': 'hello world'}  # will be the new status



import copy

import pytest

from kopf._cogs.structs.bodies import RawBody
from kopf._cogs.structs.patches import Patch


def test_error_when_no_body():
    patch = Patch()
    patch['xyz'] = 123
    with pytest.raises(ValueError, match="Cannot build a JSON-patch without the original body"):
        patch.as_json_patch()


def test_no_error_when_no_body_is_needed():
    patch = Patch()
    ops = patch.as_json_patch()
    assert ops == []


def test_body_argument_overrides_original():
    body1 = {'abc': 456}  # leads to ops=remove(abc) & op=add(xyz)
    body2 = {'xyz': 789}  # leads to op=replace(xyz) only
    patch = Patch(body=body1)
    patch['abc'] = None
    patch['xyz'] = 123
    ops = patch.as_json_patch(body2)
    assert ops == [
        {'op': 'replace', 'path': '/xyz', 'value': 123},
    ]


def test_noop_empty_patch():
    body = {'abc': 456}
    patch = Patch(body=body)
    ops = patch.as_json_patch()
    assert ops == []


def test_noop_merge_of_empty_dict():
    body = {'xyz': {'abc': 456}}
    patch = Patch(body=body)
    patch['xyz'] = {}  # nothing to merge here
    ops = patch.as_json_patch()
    assert ops == []


def test_noop_replacement_of_the_key():
    body = {'xyz': 123}
    patch = Patch(body=body)
    patch['xyz'] = 123
    ops = patch.as_json_patch()
    assert ops == []


def test_noop_removal_of_absent_key():
    body = {'abc': 456}
    patch = Patch(body=body)
    patch['xyz'] = None
    ops = patch.as_json_patch()
    assert ops == []


def test_addition_of_the_key():
    body = {'abc': 456}
    patch = Patch(body=body)
    patch['xyz'] = 123
    ops = patch.as_json_patch()
    assert ops == [
        {'op': 'add', 'path': '/xyz', 'value': 123},
    ]


def test_replacement_of_the_key():
    body = {'xyz': 456}
    patch = Patch(body=body)
    patch['xyz'] = 123
    ops = patch.as_json_patch()
    assert ops == [
        {'op': 'replace', 'path': '/xyz', 'value': 123},
    ]


def test_removal_of_the_key():
    body = {'xyz': 456}
    patch = Patch(body=body)
    patch['xyz'] = None
    ops = patch.as_json_patch()
    assert ops == [
        {'op': 'remove', 'path': '/xyz'},
    ]


def test_addition_of_the_subkey():
    body = {'xyz': {'def': 456}}
    patch = Patch(body=body)
    patch['xyz'] = {'abc': 123}
    ops = patch.as_json_patch()
    assert ops == [
        {'op': 'add', 'path': '/xyz/abc', 'value': 123},
    ]


def test_replacement_of_the_subkey():
    body = {'xyz': {'abc': 456}}
    patch = Patch(body=body)
    patch['xyz'] = {'abc': 123}
    ops = patch.as_json_patch()
    assert ops == [
        {'op': 'replace', 'path': '/xyz/abc', 'value': 123},
    ]


def test_addition_of_key_with_new_parent():
    body = {}
    patch = Patch(body=body)
    patch['xyz'] = {'abc': 123}
    ops = patch.as_json_patch()
    assert ops == [
        {'op': 'add', 'path': '/xyz', 'value': {'abc': 123}},
    ]


def test_addition_of_the_sub_subkey_with_existing_parent():
    body = {'xyz': {'uvw': 123}}
    patch = Patch(body=body)
    patch['xyz'] = {'abc': {'def': {'ghi': 456}}}
    ops = patch.as_json_patch()
    assert ops == [
        {'op': 'add', 'path': '/xyz/abc', 'value': {'def': {'ghi': 456}}},
    ]


def test_removal_of_the_subkey_and_remaining_parent():
    body = {'xyz': {'abc': 456, 'other': '...'}}
    patch = Patch(body=body)
    patch['xyz'] = {'abc': None}
    ops = patch.as_json_patch()
    assert ops == [
        {'op': 'remove', 'path': '/xyz/abc'},
    ]


def test_removal_of_the_subkey_and_emptied_parent():
    body = {'xyz': {'abc': 456}}
    patch = Patch(body=body)
    patch['xyz'] = {'abc': None}
    ops = patch.as_json_patch()
    assert ops == [
        {'op': 'remove', 'path': '/xyz'},
    ]


def test_addition_of_list_value():
    body = {'abc': 456}
    patch = Patch(body=body)
    patch['xyz'] = [1, 2, 3]
    ops = patch.as_json_patch()
    assert ops == [
        {'op': 'add', 'path': '/xyz', 'value': [1, 2, 3]},
    ]


def test_replacement_of_list_value():
    body = {'xyz': [1, 2, 3]}
    patch = Patch(body=body)
    patch['xyz'] = [4, 5]
    ops = patch.as_json_patch()
    # The jsonpatch library diffs lists element-by-element, not as a whole.
    assert sorted(ops, key=lambda op: op['path']) == [
        {'op': 'replace', 'path': '/xyz/0', 'value': 4},
        {'op': 'replace', 'path': '/xyz/1', 'value': 5},
        {'op': 'remove', 'path': '/xyz/2'},
    ]


def test_multiple_operations():
    body = {'existing': 1, 'toremove': 2}
    patch = Patch(body=body)
    patch['existing'] = 99
    patch['toremove'] = None
    patch['newkey'] = 'hello'
    ops = patch.as_json_patch()
    assert sorted(ops, key=lambda op: op['path']) == [
        {'op': 'replace', 'path': '/existing', 'value': 99},
        {'op': 'add', 'path': '/newkey', 'value': 'hello'},
        {'op': 'remove', 'path': '/toremove'},
    ]


def test_fn_only():
    body = {'items': [1, 2]}
    patch = Patch(body=body, fns=[lambda b: b['items'].append(3)])
    ops = patch.as_json_patch()
    assert ops == [
        {'op': 'add', 'path': '/items/2', 'value': 3},
    ]


def test_fn_combined_with_merge_patch():
    body = {'items': [1, 2], 'label': 'old'}
    patch = Patch(body=body, fns=[lambda b: b['items'].append(3)])
    patch['label'] = 'new'
    ops = patch.as_json_patch()
    assert sorted(ops, key=lambda op: op['path']) == [
        {'op': 'add', 'path': '/items/2', 'value': 3},
        {'op': 'replace', 'path': '/label', 'value': 'new'},
    ]


def test_fn_applied_strictly_after_merges():

    def increment(body: RawBody) -> None:
        body['xyz'] += 1

    body = {'xyz': 100}
    patch = Patch(body=body, fns=[increment])
    patch['xyz'] = 200
    ops = patch.as_json_patch()
    assert sorted(ops, key=lambda op: op['path']) == [
        {'op': 'replace', 'path': '/xyz', 'value': 201},
    ]


def test_escaping_of_key():
    body = {'~xyz/test': {'abc': '...', 'other': '...'}}
    patch = Patch(body=body)
    patch['~xyz/test'] = {'abc': None}
    ops = patch.as_json_patch()
    assert ops == [
        {'op': 'remove', 'path': '/~0xyz~1test/abc'}
    ]


def test_recursive_escape_of_key():
    body = {'x/y/~z': {'a/b/~0c': '...', 'other': '...'}}
    patch = Patch(body=body)
    patch['x/y/~z'] = {'a/b/~0c': None}
    ops = patch.as_json_patch()
    assert ops == [
        {'op': 'remove', 'path': '/x~1y~1~0z/a~1b~1~00c'},
    ]


def test_does_not_mutate_original_body():
    body = {'spec': {'x': 'original'}}
    patch = Patch(body=body)
    patch['spec'] = {'x': 'modified'}
    body_before = copy.deepcopy(body)
    patch.as_json_patch()
    assert body == body_before


def test_does_not_mutate_original_body_with_fns():
    body = {'spec': {'x': 'original'}}
    patch = Patch(body=body, fns=[lambda b: b.setdefault('extra', 'added')])
    patch['spec'] = {'x': 'modified'}
    body_before = copy.deepcopy(body)
    patch.as_json_patch()
    assert body == body_before



import pytest

from kopf._cogs.structs.dicts import MappingView, MutableMappingView, ReplaceableMappingView

all_classes = pytest.mark.parametrize('cls', [
    MappingView, MutableMappingView, ReplaceableMappingView])


@all_classes
def test_creation_with_source(cls):
    cls({})


@all_classes
def test_creation_without_source(cls):
    with pytest.raises(TypeError):
        cls()


@all_classes
def test_access_for_root(cls):
    data = {'field': 'value'}
    view = cls(data)
    value = view['field']
    assert value == 'value'


@all_classes
def test_access_for_path(cls):
    data = {'spec': {'field': 'value'}}
    view = cls(data, 'spec')
    value = view['field']
    assert value == 'value'


@all_classes
def test_unexistent_key_for_root(cls):
    data = {}
    view = cls(data)
    with pytest.raises(KeyError):
        view['field']


@all_classes
def test_unexistent_key_for_path(cls):
    data = {}
    view = cls(data, 'spec')
    with pytest.raises(KeyError):
        view['field']


@all_classes
def test_live_access_for_root(cls):
    data = {}
    view = cls(data)
    data['field'] = 'value'
    value = view['field']
    assert value == 'value'


@all_classes
def test_live_access_for_path(cls):
    data = {}
    view = cls(data, 'spec')
    data['spec'] = {'field': 'value'}
    value = view['field']
    assert value == 'value'


@all_classes
def test_length_for_root(cls):
    data = {'field': 'value'}
    view = cls(data)
    assert len(view) == 1


@all_classes
def test_length_for_path(cls):
    data = {'spec': {'field': 'value'}, 'extra-key': None}
    view = cls(data, 'spec')
    assert len(view) == 1


@all_classes
def test_bool_true_for_root(cls):
    data = {'field': 'value'}
    view = cls(data)
    assert bool(view)


@all_classes
def test_bool_false_for_root(cls):
    data = {}
    view = cls(data)
    assert not bool(view)


@all_classes
def test_bool_true_for_path(cls):
    data = {'spec': {'field': 'value'}}
    view = cls(data, 'spec')
    assert bool(view)


@all_classes
def test_bool_false_for_path(cls):
    data = {}
    view = cls(data, 'spec')
    assert not bool(view)


@all_classes
def test_keys_for_root(cls):
    data = {'field': 'value'}
    view = cls(data)
    assert list(view.keys()) == ['field']


@all_classes
def test_keys_for_path(cls):
    data = {'spec': {'field': 'value'}}
    view = cls(data, 'spec')
    assert list(view.keys()) == ['field']


@all_classes
def test_values_for_root(cls):
    data = {'field': 'value'}
    view = cls(data)
    assert list(view.values()) == ['value']


@all_classes
def test_values_for_path(cls):
    data = {'spec': {'field': 'value'}}
    view = cls(data, 'spec')
    assert list(view.values()) == ['value']


@all_classes
def test_items_for_root(cls):
    data = {'field': 'value'}
    view = cls(data)
    assert list(view.items()) == [('field', 'value')]


@all_classes
def test_items_for_path(cls):
    data = {'spec': {'field': 'value'}}
    view = cls(data, 'spec')
    assert list(view.items()) == [('field', 'value')]


@all_classes
def test_dict_for_root(cls):
    data = {'field': 'value'}
    view = cls(data)
    assert dict(view) == {'field': 'value'}


@all_classes
def test_dict_for_path(cls):
    data = {'spec': {'field': 'value'}}
    view = cls(data, 'spec')
    assert dict(view) == {'field': 'value'}


def test_update_for_root():
    data = {'field': 'value'}
    view = MutableMappingView(data)
    view['field'] = 'new-value'
    assert data == {'field': 'new-value'}


def test_update_for_path():
    data = {'spec': {'field': 'value'}}
    view = MutableMappingView(data, 'spec')
    view['field'] = 'new-value'
    assert data == {'spec': {'field': 'new-value'}}


def test_delete_for_root():
    data = {'field': 'value'}
    view = MutableMappingView(data)
    del view['field']
    assert data == {}


def test_delete_for_path():
    data = {'spec': {'field': 'value'}}
    view = MutableMappingView(data, 'spec')
    del view['field']
    assert data == {'spec': {}}


def test_chain_creation():
    data = {}
    view = MutableMappingView(data, 'spec')
    view['field'] = 'new-value'
    assert data == {'spec': {'field': 'new-value'}}


def test_replacing_from_another_view():
    data1 = {'field1': 'value1'}
    data2 = {'field2': 'value2'}
    view1 = ReplaceableMappingView(data1)
    view2 = ReplaceableMappingView(data2)
    view1._replace_from(view2)
    assert view1._src is not view2
    assert view1._src is view2._src
    assert dict(view1) == {'field2': 'value2'}


def test_replacing_with_another_view():
    data1 = {'field1': 'value1'}
    data2 = {'field2': 'value2'}
    view1 = ReplaceableMappingView(data1)
    view2 = ReplaceableMappingView(data2)
    view1._replace_with(view2)
    assert view1._src is view2
    assert dict(view1) == {'field2': 'value2'}


def test_replacing_with_regular_dict():
    data1 = {'field1': 'value1'}
    data2 = {'field2': 'value2'}
    view1 = ReplaceableMappingView(data1)
    view1._replace_with(data2)
    assert view1._src is data2
    assert dict(view1) == {'field2': 'value2'}



"""
Only the tests from the K8s client library (simulated) to the event stream.

Excluded: the watch-stream consumption and queue multiplexing routines
(see ``tests_queueing.py``).

Used for internal control that the event multiplexing works are intended.
If the intentions change, the tests should be rewritten.
They are NOT part of the public interface of the framework.
"""
import asyncio

import aiohttp
import pytest

from kopf._cogs.clients.watching import Bookmark, WatchingError, continuous_watch

STREAM_WITH_ERROR_410GONE_ONLY = (
    {'type': 'ERROR', 'object': {'code': 410}},
)
STREAM_WITH_NORMAL_EVENTS = (
    {'type': 'ADDED', 'object': {'spec': 'a'}},
    {'type': 'ADDED', 'object': {'spec': 'b'}},
)
STREAM_WITH_UNKNOWN_EVENT = (
    {'type': 'ADDED', 'object': {'spec': 'a'}},
    {'type': 'UNKNOWN', 'object': {}},
    {'type': 'ADDED', 'object': {'spec': 'b'}},
)
STREAM_WITH_ERROR_410GONE = (
    {'type': 'ADDED', 'object': {'spec': 'a'}},
    {'type': 'ERROR', 'object': {'code': 410}},
    {'type': 'ADDED', 'object': {'spec': 'b'}},
)
STREAM_WITH_ERROR_CODE = (
    {'type': 'ADDED', 'object': {'spec': 'a'}},
    {'type': 'ERROR', 'object': {'code': 666}},
    {'type': 'ADDED', 'object': {'spec': 'b'}},
)
EOS = ({'type': 'ERROR', 'object': {'code': 410}},)


@pytest.fixture(autouse=True)
def _stubs(kmock, resource):
    # The watch-stream makes an initial listing, so ensure there is a dummy response instead of 404.
    # Also ensure the continuous watcher terminates in the end (one "410 Gone" guarantees this).
    (kmock['list', resource] ** -60) << {'items': []}
    (kmock['watch', resource] ** -60) << STREAM_WITH_ERROR_410GONE_ONLY


class SampleException(Exception):
    pass


async def test_empty_stream_yields_nothing(kmock, settings, resource, namespace):

    events = []
    async for event in continuous_watch(settings=settings,
                                        resource=resource,
                                        namespace=namespace,
                                        operator_pause_waiter=asyncio.Future()):
        events.append(event)

    assert len(events) == 1
    assert events[0] == Bookmark.LISTED


async def test_event_stream_yields_everything(kmock, settings, resource, namespace):
    kmock['watch', resource, kmock.namespace(namespace)] << STREAM_WITH_NORMAL_EVENTS << EOS

    events = []
    async for event in continuous_watch(settings=settings,
                                        resource=resource,
                                        namespace=namespace,
                                        operator_pause_waiter=asyncio.Future()):
        events.append(event)

    assert len(events) == 3
    assert events[0] == Bookmark.LISTED
    assert events[1]['object']['spec'] == 'a'
    assert events[2]['object']['spec'] == 'b'


async def test_unknown_event_type_ignored(kmock, settings, resource, namespace, assert_logs):
    kmock['watch', resource, kmock.namespace(namespace)] << STREAM_WITH_UNKNOWN_EVENT << EOS

    events = []
    async for event in continuous_watch(settings=settings,
                                        resource=resource,
                                        namespace=namespace,
                                        operator_pause_waiter=asyncio.Future()):
        events.append(event)

    assert len(events) == 3
    assert events[0] == Bookmark.LISTED
    assert events[1]['object']['spec'] == 'a'
    assert events[2]['object']['spec'] == 'b'
    assert_logs(["Ignoring an unsupported event type"])
    assert_logs(["UNKNOWN"])


async def test_error_410gone_exits_normally(kmock, settings, resource, namespace, assert_logs):
    kmock['watch', resource, kmock.namespace(namespace)] << STREAM_WITH_ERROR_410GONE << EOS

    events = []
    async for event in continuous_watch(settings=settings,
                                        resource=resource,
                                        namespace=namespace,
                                        operator_pause_waiter=asyncio.Future()):
        events.append(event)

    assert len(events) == 2
    assert events[0] == Bookmark.LISTED
    assert events[1]['object']['spec'] == 'a'
    assert_logs(["Restarting the watch-stream"])


async def test_unknown_error_raises_exception(kmock, settings, resource, namespace):
    kmock['watch', resource, kmock.namespace(namespace)] << STREAM_WITH_ERROR_CODE << EOS

    events = []
    with pytest.raises(WatchingError) as e:
        async for event in continuous_watch(settings=settings,
                                            resource=resource,
                                            namespace=namespace,
                                            operator_pause_waiter=asyncio.Future()):
            events.append(event)

    assert len(events) == 2
    assert events[0] == Bookmark.LISTED
    assert events[1]['object']['spec'] == 'a'
    assert '666' in str(e.value)


async def test_exception_escalates(kmock, settings, resource, namespace, enforced_session, mocker):
    enforced_session.request = mocker.Mock(side_effect=SampleException())
    kmock['watch', resource, kmock.namespace(namespace)] << ()

    events = []
    with pytest.raises(SampleException):
        async for event in continuous_watch(settings=settings,
                                            resource=resource,
                                            namespace=namespace,
                                            operator_pause_waiter=asyncio.Future()):
            events.append(event)

    assert len(events) == 0


# See: See: https://github.com/zalando-incubator/kopf/issues/275
async def test_long_line_parsing(kmock, settings, resource, namespace):
    kmock['watch', resource, kmock.namespace(namespace)] << (
        {'type': 'ADDED', 'object': {'spec': {'field': 'x'}}},
        {'type': 'ADDED', 'object': {'spec': {'field': 'y' * (2 * 1024 * 1024)}}},
        {'type': 'ADDED', 'object': {'spec': {'field': 'z' * (4 * 1024 * 1024)}}},
    ) << EOS

    events = []
    async for event in continuous_watch(settings=settings,
                                        resource=resource,
                                        namespace=namespace,
                                        operator_pause_waiter=asyncio.Future()):
        events.append(event)

    assert len(events) == 4
    assert events[0] == Bookmark.LISTED
    assert len(events[1]['object']['spec']['field']) == 1
    assert len(events[2]['object']['spec']['field']) == 2 * 1024 * 1024
    assert len(events[3]['object']['spec']['field']) == 4 * 1024 * 1024

@pytest.mark.parametrize("connection_error",
    [
        aiohttp.ClientConnectionError,
        aiohttp.ClientPayloadError,
        asyncio.TimeoutError
    ]
)
async def test_list_objs_connection_errors_are_caught(
        settings, resource, namespace, enforced_session, mocker, connection_error):
    enforced_session.request = mocker.Mock(side_effect=connection_error())

    events = []
    async for event in continuous_watch(settings=settings,
                                            resource=resource,
                                            namespace=namespace,
                                            operator_pause_waiter=asyncio.Future()):
        events.append(event)

    assert len(events) == 0



import copy

import pytest

import kopf

forced_mode = pytest.mark.parametrize('forcedness', [
    pytest.param(dict(forced=True), id='forcedTrue'),
])
non_forced_mode = pytest.mark.parametrize('forcedness', [
    pytest.param(dict(forced=False), id='forcedFalse'),
    pytest.param(dict(), id='forcedAbsent'),
])
any_forced_mode = pytest.mark.parametrize('forcedness', [
    pytest.param(dict(forced=True), id='forcedTrue'),
    pytest.param(dict(forced=False), id='forcedFalse'),
    pytest.param(dict(), id='forcedAbsent'),
])

obj1_with_namespace = pytest.mark.parametrize('obj1', [
    pytest.param({'metadata': {'namespace': 'a'}}, id='withnamespace'),
])
obj2_with_namespace = pytest.mark.parametrize('obj2', [
    pytest.param({'metadata': {'namespace': 'a'}}, id='withnamespace'),
])
obj1_without_namespace = pytest.mark.parametrize('obj1', [
    pytest.param({}, id='withoutmetadata'),
    pytest.param({'metadata': {}}, id='withoutnamespace'),
])
obj2_without_namespace = pytest.mark.parametrize('obj2', [
    pytest.param({}, id='withoutmetadata'),
    pytest.param({'metadata': {}}, id='withoutnamespace'),
])


# In the NON-FORCED mode, the EXISTING namespaces are preserved.
@obj1_with_namespace
@non_forced_mode
def test_preserved_namespace_of_dict(forcedness, obj1):
    obj1 = copy.deepcopy(obj1)
    kopf.adjust_namespace(obj1, namespace='provided-namespace', **forcedness)
    assert 'namespace' in obj1['metadata']
    assert obj1['metadata']['namespace'] != 'provided-namespace'


@obj2_with_namespace
@obj1_with_namespace
@non_forced_mode
def test_preserved_namespaces_of_dicts(forcedness, multicls, obj1, obj2):
    obj1, obj2 = copy.deepcopy(obj1), copy.deepcopy(obj2)
    objs = multicls([obj1, obj2])
    kopf.adjust_namespace(objs, namespace='provided-namespace', **forcedness)
    assert 'namespace' in obj1['metadata']
    assert 'namespace' in obj2['metadata']
    assert obj1['metadata']['namespace'] != 'provided-namespace'
    assert obj2['metadata']['namespace'] != 'provided-namespace'


@obj1_with_namespace
@non_forced_mode
def test_preserved_namespace_of_pykube_object(forcedness, pykube_object, obj1):
    pykube_object.obj = copy.deepcopy(obj1)
    kopf.adjust_namespace(pykube_object, namespace='provided-namespace', **forcedness)
    assert pykube_object.obj['metadata'].get('namespace') != 'provided-namespace'


@obj1_with_namespace
@non_forced_mode
def test_preserved_namespace_of_kubernetes_model(forcedness, kubernetes_model, obj1):
    kubernetes_model.metadata.namespace = obj1.get('metadata', {}).get('namespace')
    kopf.adjust_namespace(kubernetes_model, namespace='provided-namespace', **forcedness)
    assert kubernetes_model.metadata.namespace != 'provided-namespace'


@obj1_with_namespace
@non_forced_mode
def test_preserved_namespace_of_kubernetes_asyncio_model(forcedness, kubernetes_asyncio_model, obj1):
    kubernetes_asyncio_model.metadata.namespace = obj1.get('metadata', {}).get('namespace')
    kopf.adjust_namespace(kubernetes_asyncio_model, namespace='provided-namespace', **forcedness)
    assert kubernetes_asyncio_model.metadata.namespace != 'provided-namespace'


#
# In the FORCED mode, the EXISTING namespaces are overwritten.
#
@obj1_with_namespace
@forced_mode
def test_overwriting_of_namespace_of_dict(forcedness, obj1):
    obj1 = copy.deepcopy(obj1)
    kopf.adjust_namespace(obj1, namespace='provided-namespace', **forcedness)
    assert 'namespace' in obj1['metadata']
    assert obj1['metadata']['namespace'] == 'provided-namespace'


@obj2_with_namespace
@obj1_with_namespace
@forced_mode
def test_overwriting_of_namespaces_of_dicts(forcedness, multicls, obj1, obj2):
    obj1, obj2 = copy.deepcopy(obj1), copy.deepcopy(obj2)
    objs = multicls([obj1, obj2])
    kopf.adjust_namespace(objs, namespace='provided-namespace', **forcedness)
    assert 'namespace' in obj1['metadata']
    assert 'namespace' in obj2['metadata']
    assert obj1['metadata']['namespace'] == 'provided-namespace'
    assert obj2['metadata']['namespace'] == 'provided-namespace'


@obj1_with_namespace
@forced_mode
def test_overwriting_namespace_of_pykube_object(forcedness, pykube_object, obj1):
    pykube_object.obj = copy.deepcopy(obj1)
    kopf.adjust_namespace(pykube_object, namespace='provided-namespace', **forcedness)
    assert pykube_object.obj['metadata'].get('namespace') == 'provided-namespace'


@obj1_with_namespace
@forced_mode
def test_overwriting_namespace_of_kubernetes_model(forcedness, kubernetes_model, obj1):
    kubernetes_model.metadata.namespace = obj1.get('metadata', {}).get('namespace')
    kopf.adjust_namespace(kubernetes_model, namespace='provided-namespace', **forcedness)
    assert kubernetes_model.metadata.namespace == 'provided-namespace'


@obj1_with_namespace
@forced_mode
def test_overwriting_namespace_of_kubernetes_asyncio_model(forcedness, kubernetes_asyncio_model, obj1):
    kubernetes_asyncio_model.metadata.namespace = obj1.get('metadata', {}).get('namespace')
    kopf.adjust_namespace(kubernetes_asyncio_model, namespace='provided-namespace', **forcedness)
    assert kubernetes_asyncio_model.metadata.namespace == 'provided-namespace'


#
# When namespaces are ABSENT, they are added regardless of the forced mode.
#
@obj1_without_namespace
@any_forced_mode
def test_assignment_of_namespace_of_dict(forcedness, obj1):
    obj1 = copy.deepcopy(obj1)
    kopf.adjust_namespace(obj1, namespace='provided-namespace', **forcedness)
    assert 'namespace' in obj1['metadata']
    assert obj1['metadata']['namespace'] == 'provided-namespace'


@obj2_without_namespace
@obj1_without_namespace
@any_forced_mode
def test_assignment_of_namespaces_of_dicts(forcedness, multicls, obj1, obj2):
    obj1, obj2 = copy.deepcopy(obj1), copy.deepcopy(obj2)
    objs = multicls([obj1, obj2])
    kopf.adjust_namespace(objs, namespace='provided-namespace', **forcedness)
    assert 'namespace' in obj1['metadata']
    assert 'namespace' in obj2['metadata']
    assert obj1['metadata']['namespace'] == 'provided-namespace'
    assert obj2['metadata']['namespace'] == 'provided-namespace'


@obj1_without_namespace
@any_forced_mode
def test_assignment_namespace_of_pykube_object(forcedness, pykube_object, obj1):
    pykube_object.obj = copy.deepcopy(obj1)
    kopf.adjust_namespace(pykube_object, namespace='provided-namespace', **forcedness)
    assert pykube_object.obj['metadata'].get('namespace') == 'provided-namespace'


@any_forced_mode
def test_assignment_namespace_of_kubernetes_model(forcedness, kubernetes_model):
    kubernetes_model.metadata = None
    kopf.adjust_namespace(kubernetes_model, namespace='provided-namespace', **forcedness)
    assert kubernetes_model.metadata.namespace == 'provided-namespace'


@any_forced_mode
def test_assignment_namespace_of_kubernetes_asyncio_model(forcedness, kubernetes_asyncio_model):
    kubernetes_asyncio_model.metadata = None
    kopf.adjust_namespace(kubernetes_asyncio_model, namespace='provided-namespace', **forcedness)
    assert kubernetes_asyncio_model.metadata.namespace == 'provided-namespace'



import asyncio

import pytest

import kopf
from kopf._cogs.structs.ephemera import Memo
from kopf._core.engines.indexing import OperatorIndexers
from kopf._core.intents.causes import Reason
from kopf._core.reactor.inventory import ResourceMemories
from kopf._core.reactor.processing import process_resource_event

LAST_SEEN_ANNOTATION = 'kopf.zalando.org/last-handled-configuration'

EVENT_TYPES = [None, 'ADDED', 'MODIFIED', 'DELETED']
EVENT_TYPES_WHEN_EXISTS = [None, 'ADDED', 'MODIFIED']


@pytest.mark.parametrize('event_type', EVENT_TYPES_WHEN_EXISTS)
async def test_create(registry, settings, handlers, resource, cause_mock, event_type,
                      assert_logs, k8s_mocked):
    cause_mock.reason = Reason.CREATE

    settings.posting.loggers = True
    event_queue = asyncio.Queue()
    await process_resource_event(
        lifecycle=kopf.lifecycles.all_at_once,
        registry=registry,
        settings=settings,
        resource=resource,
        indexers=OperatorIndexers(),
        memories=ResourceMemories(),
        memobase=Memo(),
        raw_event={'type': event_type, 'object': {}},
        event_queue=event_queue,
    )

    assert handlers.create_mock.call_count == 1
    assert not handlers.update_mock.called
    assert not handlers.delete_mock.called

    assert k8s_mocked.patch.call_count == 1
    assert not event_queue.empty()

    patch = k8s_mocked.patch.call_args_list[0].kwargs['payload']
    assert 'metadata' in patch
    assert 'annotations' in patch['metadata']
    assert LAST_SEEN_ANNOTATION in patch['metadata']['annotations']

    assert_logs([
        "Creation is in progress:",
        "Handler 'create_fn' is invoked",
        "Handler 'create_fn' succeeded",
        "Creation is processed:",
        "Merge-patching",
    ])


@pytest.mark.parametrize('event_type', EVENT_TYPES_WHEN_EXISTS)
async def test_update(registry, settings, handlers, resource, cause_mock, event_type,
                      assert_logs, k8s_mocked):
    cause_mock.reason = Reason.UPDATE

    settings.posting.loggers = True
    event_queue = asyncio.Queue()
    await process_resource_event(
        lifecycle=kopf.lifecycles.all_at_once,
        registry=registry,
        settings=settings,
        resource=resource,
        indexers=OperatorIndexers(),
        memories=ResourceMemories(),
        memobase=Memo(),
        raw_event={'type': event_type, 'object': {}},
        event_queue=event_queue,
    )

    assert not handlers.create_mock.called
    assert handlers.update_mock.call_count == 1
    assert not handlers.delete_mock.called

    assert k8s_mocked.patch.call_count == 1
    assert not event_queue.empty()

    patch = k8s_mocked.patch.call_args_list[0].kwargs['payload']
    assert 'metadata' in patch
    assert 'annotations' in patch['metadata']
    assert LAST_SEEN_ANNOTATION in patch['metadata']['annotations']

    assert_logs([
        "Updating is in progress:",
        "Handler 'update_fn' is invoked",
        "Handler 'update_fn' succeeded",
        "Updating is processed:",
        "Merge-patching",
    ])


@pytest.mark.parametrize('event_type', EVENT_TYPES_WHEN_EXISTS)
async def test_delete(registry, settings, handlers, resource, cause_mock, event_type,
                      assert_logs, k8s_mocked):
    cause_mock.reason = Reason.DELETE
    finalizer = settings.persistence.finalizer
    event_body = {'metadata': {'deletionTimestamp': '...', 'finalizers': [finalizer]}}

    settings.posting.loggers = True
    event_queue = asyncio.Queue()
    await process_resource_event(
        lifecycle=kopf.lifecycles.all_at_once,
        registry=registry,
        settings=settings,
        resource=resource,
        indexers=OperatorIndexers(),
        memories=ResourceMemories(),
        memobase=Memo(),
        raw_event={'type': event_type, 'object': event_body},
        event_queue=event_queue,
    )

    assert not handlers.create_mock.called
    assert not handlers.update_mock.called
    assert handlers.delete_mock.call_count == 1

    assert k8s_mocked.patch.call_count == 2  # (1) annotations, (2) finalizers.
    assert not event_queue.empty()

    assert_logs([
        "Deletion is in progress:",
        "Handler 'delete_fn' is invoked",
        "Handler 'delete_fn' succeeded",
        "Deletion is processed:",
        "Removing the finalizer",
        "Merge-patching",
    ])


#
# Informational causes: just log, and do nothing else.
#

@pytest.mark.parametrize('consistency_time', [0, 100], ids=['now', 'future'])
@pytest.mark.parametrize('event_type', EVENT_TYPES)
async def test_gone(registry, settings, handlers, resource, cause_mock, event_type,
                    assert_logs, k8s_mocked, consistency_time, looptime):
    cause_mock.reason = Reason.GONE

    settings.posting.loggers = True
    event_queue = asyncio.Queue()
    await process_resource_event(
        lifecycle=kopf.lifecycles.all_at_once,
        registry=registry,
        settings=settings,
        resource=resource,
        indexers=OperatorIndexers(),
        memories=ResourceMemories(),
        memobase=Memo(),
        raw_event={'type': event_type, 'object': {}},
        event_queue=event_queue,
        consistency_time=consistency_time,
    )

    assert not handlers.create_mock.called
    assert not handlers.update_mock.called
    assert not handlers.delete_mock.called

    assert not k8s_mocked.patch.called
    assert event_queue.empty()

    assert looptime == 0
    assert_logs([
        "Deleted, really deleted",
    ])


# Happens on the removal of the very last finalizer on DELETED events (it arrives as not removed).
async def test_gone_with_finalizer_not_removed_on_deleted_event(
        registry, settings, handlers, resource, cause_mock,
        assert_logs, k8s_mocked):
    cause_mock.reason = Reason.GONE
    finalizer = settings.persistence.finalizer
    event_body = {'metadata': {'deletionTimestamp': '...', 'finalizers': [finalizer]}}

    event_queue = asyncio.Queue()
    await process_resource_event(
        lifecycle=kopf.lifecycles.all_at_once,
        registry=registry,
        settings=settings,
        resource=resource,
        indexers=OperatorIndexers(),
        memories=ResourceMemories(),
        memobase=Memo(),
        raw_event={'type': 'DELETED', 'object': event_body},
        event_queue=event_queue,
    )

    assert not handlers.create_mock.called
    assert not handlers.update_mock.called
    assert not handlers.delete_mock.called

    assert not k8s_mocked.patch.called
    assert event_queue.empty()

    assert_logs([
        "Deleted, really deleted",
    ], prohibited=[
        "Removing the finalizer",
    ])


@pytest.mark.parametrize('event_type', EVENT_TYPES)
async def test_free(registry, settings, handlers, resource, cause_mock, event_type,
                    assert_logs, k8s_mocked):
    cause_mock.reason = Reason.FREE

    settings.posting.loggers = True
    event_queue = asyncio.Queue()
    await process_resource_event(
        lifecycle=kopf.lifecycles.all_at_once,
        registry=registry,
        settings=settings,
        resource=resource,
        indexers=OperatorIndexers(),
        memories=ResourceMemories(),
        memobase=Memo(),
        raw_event={'type': event_type, 'object': {}},
        event_queue=event_queue,
    )

    assert not handlers.create_mock.called
    assert not handlers.update_mock.called
    assert not handlers.delete_mock.called
    assert not k8s_mocked.patch.called
    assert event_queue.empty()

    assert_logs([
        "Deletion, but we are done with it",
    ])


@pytest.mark.parametrize('event_type', EVENT_TYPES)
async def test_noop(registry, settings, handlers, resource, cause_mock, event_type,
                    assert_logs, k8s_mocked):
    cause_mock.reason = Reason.NOOP

    settings.posting.loggers = True
    event_queue = asyncio.Queue()
    await process_resource_event(
        lifecycle=kopf.lifecycles.all_at_once,
        registry=registry,
        settings=settings,
        resource=resource,
        indexers=OperatorIndexers(),
        memories=ResourceMemories(),
        memobase=Memo(),
        raw_event={'type': event_type, 'object': {}},
        event_queue=event_queue,
    )

    assert not handlers.create_mock.called
    assert not handlers.update_mock.called
    assert not handlers.delete_mock.called
    assert not k8s_mocked.patch.called
    assert event_queue.empty()

    assert_logs([
        "Something has changed, but we are not interested",
    ])
